# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example of excluding certain test records from the output callbacks.

In this case, we exclude tests which were aborted before a DUT ID was set, since
they are unlikely to contain any useful information. Note that "abort" refers to
a KeyboardInterrupt. If any other error occurs before the DUT ID is set, those
records are not excluded, since they may be relevant for debugging.

It may make sense to implement this check if your hardware tests follow the
common pattern of waiting for the DUT ID to be entered via a prompt at the test
start.
"""

import openhtf as htf
from openhtf.output.callbacks import json_factory
from openhtf.core import test_record
from openhtf.plugs import user_input
from openhtf.util import console_output

DEFAULT_DUT_ID = '<UNSET_DUT_ID>'


class CustomOutputToJSON(json_factory.OutputToJSON):

  def __call__(self, record):
    if (record.outcome == test_record.Outcome.ABORTED
        and record.dut_id == DEFAULT_DUT_ID):
      console_output.cli_print(
          'Test was aborted at test start. Skipping output to JSON.')
    else:
      console_output.cli_print('Outputting test record to JSON.')
      super(CustomOutputToJSON, self).__call__(record)


@htf.plug(user=user_input.UserInput)
def HelloWorldPhase(test, user):
  test.logger.info('Hello World!')
  user.prompt('The DUT ID is `%s`. Press enter to continue.' %
              test.test_record.dut_id)


def main():
  test = htf.Test(HelloWorldPhase)
  test.configure(default_dut_id=DEFAULT_DUT_ID)
  test.add_output_callbacks(
      CustomOutputToJSON('./{dut_id}.hello_world.json', indent=2))
  test.execute(test_start=user_input.prompt_for_test_start())


if __name__ == '__main__':
  main()



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Base interface for communicating with USB devices.

This module provides the base classes required to support interfacing with USB
devices through a few different types of USB handles.  All handles implement the
UsbHandle interface, which contains Open, Close, Read, and Write functionality.
A UsbHandle object represents a single USB Interface, *not* an entire device.
"""

from future.utils import with_metaclass
import abc
import functools
import logging

from openhtf.plugs.usb import usb_exceptions

DEFAULT_TIMEOUT_MS = 5000
FLUSH_READ_SIZE = 1024 * 64
_LOG = logging.getLogger(__name__)


def requires_open_handle(method):  # pylint: disable=invalid-name
  """Decorator to ensure a handle is open for certain methods.

  Subclasses should decorate their Read() and Write() with this rather than
  checking their own internal state, keeping all "is this handle open" logic
  in is_closed().

  Args:
    method: A class method on a subclass of UsbHandle

  Raises:
    HandleClosedError: If this handle has been closed.

  Returns:
    A wrapper around method that ensures the handle is open before calling through
  to the wrapped method.
  """
  @functools.wraps(method)
  def wrapper_requiring_open_handle(self, *args, **kwargs):
    """The wrapper to be returned."""
    if self.is_closed():
      raise usb_exceptions.HandleClosedError()
    return method(self, *args, **kwargs)
  return wrapper_requiring_open_handle


class UsbHandle(with_metaclass(abc.ABCMeta, object)):
  """UsbHandle objects provide read/write access to USB Interfaces.

  Subclasses must implement this interface to provide actual Read/Write/Close
  functionality.  Applications using UsbHandle objects should be agnostic as to
  the specific implementation used (ie, should only use methods exposed in this
  base class), and should use the subclasses only to determine the mechanism by
  which the handle is opened.

  Read and Write perform a Bulk or Interrupt transfer, depending on the type of
  the endpoints opened, so that applications need only open the correct
  Interface and call Read or Write, without having to worry about whether it is
  a Bulk or Interrupt transfer.

  It is recommended that subclasses implement the following Open() and
  iter_open() classmethods:
    Open():
      Minimum Args:
        interface_class: USB interface_class to match
        interface_subclass: USB interface_subclass to match
        interface_protocol: USB interface_protocol to match
        serial_number: USB serial number to match
        default_timeout_ms: Default timeout to use for the returned Handle.

      Returns:
        Instance of UsbHandle for communicating with the matched device, or
      None if no matching device was found.

      Raises:
        MultipleInterfacesFoundError: If multiple interfaces were found that
      match the given criteria.

    iter_open():
      Same Args as Open(), but instead of returning a UsbHandle, yields
    UsbHandle instances for all matching connected devices.

  Attributes:
    name: Name assigned to this handle, used for logging purposes.
    serial_number: Serial number of the device this handle refers to.
  """

  def __init__(self, serial_number, name=None, default_timeout_ms=None):
    """Create a UsbHandle to refer to a particular USB Interface.

    Note that this base class implementation doesn't actually track much, that's
    up to subclasses.  We track a name that can optionally be used for more
    useful log messages, a default timeout for read/write operations, and serial
    number, as it is universally available for all USB devices.

    Args:
      serial_number: Serial number of this USB device.
      name: A name for this UsbHandle, only used for log messages.
      default_timeout_ms: Default timeout for reads/writes, in milliseconds.
    """
    self.serial_number = serial_number
    self.name = name or ''
    self._default_timeout_ms = default_timeout_ms or DEFAULT_TIMEOUT_MS

  def __del__(self):  # pylint: disable=invalid-name
    if not self.is_closed():
      _LOG.error('!!!!!USB!!!!! %s not closed!', type(self).__name__)

  def __str__(self):
    return '<%s: (%s %s)>' % (type(self).__name__, self.name,
                              self.serial_number)
  __repr__ = __str__

  def _timeout_or_default(self, timeout_ms):
    """Specify a timeout or take the default."""
    return int(timeout_ms if timeout_ms is not None
               else self._default_timeout_ms)

  def flush_buffers(self):
    """Default implementation, calls Read() until it blocks."""
    while True:
      try:
        self.read(FLUSH_READ_SIZE, timeout_ms=10)
      except usb_exceptions.LibusbWrappingError as exception:
        if exception.is_timeout():
          break
        raise

  @abc.abstractmethod
  def read(self, length, timeout_ms=None):
    """Perform a USB Read on this interface, return the result.

    Implementors should call self._EnsureOpen() first, then perform any
    implementation-specific steps necessary to execute the read.

    Args:
      length: Number of bytes to attempt to read.
      timeout_ms: Timeout for this read (in millis), if None, use default.

    Returns:
      Data read by performing a Read operation.

    Raises:
      HandleClosedError: If this handle has been closed.
      UsbReadFailedError: If there was an IO error during the read.
    """

  @abc.abstractmethod
  def write(self, data, timeout_ms=None):
    """Perform a USB Write on this interface.

    Implementors should call self._EnsureOpen() first, then perform any
    implementation-specific steps necessary to execute the write.

    Args:
      data: The data to write.
      timeout_ms: Timeout for this write (in millis), if None, use default.

    Raises:
      HandleClosedError: If this handle has been closed.
      UsbWriteFailedError: If there was an IO error during the write.
    """

  @abc.abstractmethod
  def close(self):
    """Close this handle and release any associated resources.

    Other methods must not be called after calling Close().  This method must
    not raise on error.
    """

  @abc.abstractmethod
  def is_closed(self):
    """Returns True if this handle has been closed, False otherwise."""



# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Extensible HTTP server serving the OpenHTF Angular frontend."""

import os
import threading
import time

import tornado.httpclient
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web

_SERVER_SHUTDOWN_BUFFER_S = 0.5

# The directory containing the built Angular app.
WEB_GUI = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'web_gui')
STATIC_FILES_ROOT = os.path.join(WEB_GUI, 'dist')

INDEX_TEMPLATE = 'index.html'

STATIC_FILES = (
    r'css/.*\.css',
    r'css/.*\.css.map',
    r'img/.*',
    r'js/.*\.js',
    r'js/.*\.js\.map',
    r'service-worker\.js',
)
STATIC_FILE_ROUTES = '/(%s)' % '|'.join(STATIC_FILES)


def bind_port(requested_port):
  """Bind sockets to an available port, returning sockets and the bound port."""
  sockets = tornado.netutil.bind_sockets(requested_port)

  if requested_port != 0:
    return sockets, requested_port

  # Get the actual port number.
  for s in sockets:
    host, port = s.getsockname()[:2]
    if host == '0.0.0.0':
      return sockets, port

  raise RuntimeError('Could not determine the bound port.')


class CorsRequestHandler(tornado.web.RequestHandler):
  """Base handler for resources that must be accessible from other domains."""

  def set_default_headers(self):
    self.set_header('Access-Control-Allow-Origin', '*')
    self.set_header('Access-Control-Allow-Headers', 'Content-Type')

  def options(self, **unused_kwargs):
    self.set_status(204)
    self.finish()


class DefaultHandler(CorsRequestHandler):
  """A custom default handler which allows us to enable CORS on 404s."""

  def prepare(self):
    self.set_status(404)
    self.finish()


class IndexHandler(tornado.web.RequestHandler):
  """GET endpoint for the home page."""
  config = None  # Set via with_config().

  @classmethod
  def with_config(cls, config):
    return type(cls.__name__, (cls,), {'config': config})

  def get(self):
    assert self.config is not None
    self.render(INDEX_TEMPLATE, config=self.config)


class StaticFileHandler(tornado.web.StaticFileHandler):

  @classmethod
  def get_absolute_path(cls, root, path):
    return os.path.join(root, path)

  def validate_absolute_path(self, root, abspath):
    return abspath


class TemplateLoader(tornado.template.Loader):

  def __init__(self, root_directory, **kwargs):
    super(TemplateLoader, self).__init__(root_directory, **kwargs)

  def resolve_path(self, name, parent_path=None):
    return name


class WebGuiServer(threading.Thread):
  """Serves the OpenHTF Angular frontend."""

  def __init__(self, additional_routes, port, sockets=None, static_files_root=None):
    super(WebGuiServer, self).__init__(name=type(self).__name__)
    
    if static_files_root is None:
      static_files_root = STATIC_FILES_ROOT

    # Set up routes.
    routes = [
        ('/', IndexHandler.with_config(self._get_config())),
        (STATIC_FILE_ROUTES, StaticFileHandler, {
            'path': static_files_root
        }),
    ]
    routes.extend(additional_routes)

    if sockets is None:
      sockets, self.port = bind_port(port)
    else:
      if not port:
        raise ValueError('When sockets are passed to the server, port must be '
                         'specified and nonzero.')
      self.port = port

    # Configure the Tornado application.
    self.application = tornado.web.Application(
        routes,
        default_handler_class=DefaultHandler,
        template_loader=TemplateLoader(static_files_root),
        static_path=static_files_root,)
    self.server = tornado.httpserver.HTTPServer(self.application)
    self.server.add_sockets(sockets)

  def __enter__(self):
    self.start()
    return self

  def __exit__(self, *unused_args):
    self.stop()

  def _get_config(self):
    """Override this to configure the Angular app."""
    return {}

  def run(self):
    tornado.ioloop.IOLoop.instance().start()  # Blocking IO loop.

  def stop(self):
    self.server.stop()
    ioloop = tornado.ioloop.IOLoop.instance()
    ioloop.add_timeout(time.time() + _SERVER_SHUTDOWN_BUFFER_S, ioloop.stop)



import pytz
from datetime import datetime

from openhtf.core.test_record import Outcome, PhaseOutcome
from openhtf.core.measurements import Outcome as MeasureOutcome
from openhtf.util.data import convert_to_base_types

from spintop.transforms import Transformer
from spintop.models import (
    SpintopTreeTestRecordBuilder
)

def duration_of(openhtf_record):
    return (openhtf_record['end_time_millis'] - openhtf_record['start_time_millis'])/1000.0

def create_outcome_validator(openhtf_record, outcome_cls=Outcome):
    return lambda expected_outcome: outcome_cls[openhtf_record['outcome']] == expected_outcome

class OpenHTFTestRecordTransformer(Transformer):
    
    def __call__(self, test_record):
        test_record = convert_to_base_types(test_record)
        builder = SpintopTreeTestRecordBuilder()
        
        outcome_is = create_outcome_validator(test_record)
        builder.set_top_level_information(
            start_datetime=datetime.utcfromtimestamp(test_record['start_time_millis']/1000.0).replace(tzinfo=pytz.timezone('UTC')),
            dut_id=test_record['dut_id'],
            testbench_name=test_record['metadata']['test_name'],
            duration=duration_of(test_record),
            outcome=dict(
                is_pass=outcome_is(Outcome.PASS),
                is_abort=outcome_is(Outcome.ABORTED)
            )
        )
        for phase in test_record['phases']:
            phase_outcome_is = create_outcome_validator(phase, PhaseOutcome)
            with builder.new_phase(
                name=phase['name'],
                outcome=phase_outcome_is(PhaseOutcome.PASS),
                duration=duration_of(phase),
            ) as phase_builder:
                for measure, value in phase['measurements'].items():
                    phase_builder.new_measure(
                        name=measure,
                        outcome=MeasureOutcome[value['outcome']] == MeasureOutcome.PASS,
                        value=value['measured_value']
                    )
        
        return builder.build()


import pytest
import time
import threading
import queue

from spintop_openhtf.plugs.comport import ComportInterface, IOTargetTimeout

class FakeSerial(object):
    def __init__(self):
        self.to_read = ""
        self.is_open = True
        self.in_waiting = False
        self.echo = True
        self.was_read = threading.Event()

    def close(self):
        self.is_open = False

    def read(self, *args, **kwargs):
        if self.to_read:
            value = self.to_read
            self.to_read = ""
            self.in_waiting = False
            self.was_read.set()
            return value.encode('utf8')
        else:
            time.sleep(0.5)
            return None
    
    def write(self, string):
        if self.echo:
            self.in_waiting = True
            self.to_read += string.decode('utf8')

@pytest.fixture()
def fake_serial():
    fake_serial = FakeSerial()
    return fake_serial

@pytest.fixture()
def comport(fake_serial):
    # comport = ComportInterface('/dev/serial/by-id/usb-STMicroelectronics_STM32_Virtual_ComPort_in_FS_Mode_00000000050C-if00', 115200)
    comport = ComportInterface(None, None)
    comport.open(fake_serial)
    yield comport
    comport.close()


def test_receive_no_eol(fake_serial, comport):
    comport.write('Hello')
    line = comport.next_line(timeout=2)
    assert line == 'Hello'

def test_receive_immediate_eol(fake_serial, comport):
    fake_serial.was_read.clear()
    comport.write('Hello\n')
    # no wait now (only enough time for the thread to call read)
    fake_serial.was_read.wait()
    assert comport.next_line(timeout=0.1) == 'Hello\n'

def test_multi_lines(fake_serial, comport):
    fake_serial.was_read.clear()
    LINES = ('0\n', '1\n', '2\n')

    for line in LINES:
        comport.write(line)

    fake_serial.was_read.wait()

    for line in LINES:
        assert comport.next_line(timeout=0.1) == line


def test_keep_lines(fake_serial, comport):
    comport.write('0\n')
    comport.write('1\n')
    comport.write('2\n')

    fake_serial.was_read.wait()
    time.sleep(0.1)
    comport.keep_lines(1)

    assert comport.next_line(timeout=0.1) == '2\n'
    

def test_keep_lines_0(fake_serial, comport):
    comport.write('0\n')
    comport.write('1\n')
    comport.write('2\n')

    fake_serial.was_read.wait()
    time.sleep(0.1)
    comport.keep_lines(0)

    assert comport.next_line(timeout=0.1) is None

def test_clear(fake_serial, comport):
    comport.write('0\n')
    comport.clear_lines()

    assert comport.next_line(timeout=0.1) is None

def test_message_target(fake_serial, comport):
    def write_future():
        for i in (1,2,3):
            time.sleep(0.2)
            comport.write('line%d' % i)
    
    thread = threading.Thread(target=write_future)
    thread.start()
    response = comport.message_target('hello\n', 'line3')

    assert response == 'hello\nline1line2line3'
    
def test_message_target_timeout(fake_serial, comport):
    with pytest.raises(IOTargetTimeout):
        response = comport.message_target('hello\n', 'line3', timeout=0.2)


import os
import warnings
import tornado

from tornado import web, iostream
from openhtf.output.servers.station_server import StationServer as HTFStationServer
from .file_provider import TEMPFILES_PATH

class StationServer(HTFStationServer):
    def __init__(self, file_provider, *args, **kwargs):
        super(StationServer, self).__init__(*args, **kwargs)
        
        self.file_provider = file_provider
        
        self.application.add_handlers(
            r".*",  # match any host
            [
                (
                    TEMPFILES_PATH + r"([^/]*)",
                    TemporaryFileHandler,
                    { 'file_provider': self.file_provider }
                ),
            ]
        )

            
        
class TemporaryFileHandler(tornado.web.RequestHandler):
    def initialize(self, file_provider):
        self.files = file_provider.files
    
    async def get(self, *args, **kwargs):
        try:
            filename = self.files[self.request.path]
        except KeyError:
            raise tornado.web.HTTPError(404)
        
        await self._write_file_chunks(filename)
        
    async def _write_file_chunks(self, filename):
        # chunk size to read
        chunk_size = 1024 * 1024 * 1 # 1 MiB

        with open(filename, 'rb') as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                try:
                    self.write(chunk) # write the cunk to response
                    await self.flush() # flush the current chunk to socket
                except iostream.StreamClosedError:
                    # this means the client has closed the connection
                    # so break the loop
                    break
                finally:
                    # deleting the chunk is very important because 
                    # if many clients are downloading files at the 
                    # same time, the chunks in memory will keep 
                    # increasing and will eat up the RAM
                    del chunk


import os
import sys
import inspect
import datetime
import warnings

from typing import Callable, Union

from copy import copy
from contextlib import contextmanager

import pkg_resources

import openhtf as htf

from openhtf.util import conf
from openhtf.plugs import user_input, BasePlug

import webbrowser

from ..storage import SITE_DATA_DIR
from ..callbacks.file_provider import TemporaryFileProvider
from ..callbacks.local_storage import LocalStorageOutput


try:
    import tornado, sockjs
    tornado_version = pkg_resources.get_distribution("tornado").version
    major, *rest = tornado_version.split('.')
    if int(major) > 4:
        raise ValueError('Tornado version must be <= 4.x.x')
except (ImportError, ValueError):
    warnings.warn(
        'Tornado 4.x.x not available. GUI server will not work. '
        'If you wish to install it, you may install spintop-openhtf '
        'with the [server] dependency, such as '
        'pip install spintop-openhtf[server]. Do note that this requires '
        'tornado<5.0.'
    )
else:
    from ..callbacks import station_server

from .. import (
    Test,
    # load_component_file,
    # CoverageAnalysis
)

HISTORY_BASE_PATH = os.path.join(SITE_DATA_DIR, 'openhtf-history')

DEFAULT = object()

class TestPlanError(Exception): pass

class TestSequence(object):
    """The base sequence object: defines a sequence of setup, test and teardown openhtf phases.
    
    TestSequences can be nested together in two different ways:
    
    Option 1. 
        If the sequence should be re-usable, create an explicit :class:`TestSequence` object 
        and add :meth:`~TestSequence.setup`, :meth:`~TestSequence.testcase` and :meth:`~TestSequence.teardown`
        phases using the decorator methods. Afterwards, add it to a specific TestPlan (or sub TestSequence)
        using :meth:`~TestSequence.append`:

        .. highlight:: python
        .. code-block:: python

            from spintop_openhtf import TestSequence, TestPlan

            my_sequence = TestSequence('my-sequence')

            @my_sequence.testcase('Test1')
            def do_something(test):
                pass

            # Elsewhere probably

            plan = TestPlan('my-plan')
            plan.append(my_sequence)

    Option 2.
        If the sequence is defined simply to nest phases inside the same test plan, the
        :meth:`~TestSequence.sub_sequence` simplifies the declaration and usage of a sub
        sequence. This code block is equivalent to the previous one, but does not allow
        re-usage of the sub_sequence object:
        
        .. highlight:: python
        .. code-block:: python

            from spintop_openhtf import TestPlan

            plan = TestPlan('my-plan')

            my_sequence = plan.sub_sequence('my-sequence')

            @my_sequence.testcase('Test1')
            def do_something(test):
                pass

    Since TestPlans are TestSequences, and sub sequences are also TestSequences,
    they can be infinitely nested using :meth:`~TestSequence.append` and :meth:`~TestSequence.sub_sequence`.

    Args:
        name: The name of this test node.
    """
    def __init__(self, name):
        self._setup_phases = []
        self._test_phases = []
        self._teardown_phases = []
        self.name = name
    
    def setup(self, name, **options):
        """Decorator factory for a setup function.
        
        A setup function will executed if the sequence is entered. 
        All setup functions are executed before the testcases, regardless of the order
        of declaration.

        See :meth:`~TestSequence.testcase` for usage.
        """
        return self._decorate_phase(name, self._setup_phases, options)
    
    def testcase(self, name, **options):
        """Decorator factory for a normal phase. 
        
        A testcase function is a normal openhtf phase.
        
        The options parameter is a proxy for the PhaseOptions arguments.
                
        .. highlight:: python
        .. code-block:: python
        
            my_sequence = TestSequence('Parent')
        
            @my_sequence.testcase('my-testcase-name')
            def setup_fn(test):
                (...)
        
        {phase_options_doc}
        """
        return self._decorate_phase(name, self._test_phases, options)
    
    # Add PhaseOptions documentation
    testcase.__doc__ = testcase.__doc__.format(phase_options_doc=htf.PhaseOptions.__doc__)
    
    def teardown(self, name, **options):
        """Decorator factory for a teardown phase. 

        A teardown function will always be executed if the sequence is entered, regardless
        of the outcome of normal or setup phases.

        See :meth:`~TestSequence.testcase` for usage.
        """
        return self._decorate_phase(name, self._teardown_phases, options)
    
    def plug(self, *args, **kwargs):
        """Helper method: shortcut to :func:`openhtf.plug`"""
        return htf.plugs.plug(*args, **kwargs)
    
    def measures(self, *args, **kwargs):
        """Helper method: shortcut to :func:`openhtf.measures`"""
        return htf.plugs.plug(*args, **kwargs)
    
    def sub_sequence(self, name):
        """Create new empty TestSequence and append it to this sequence.
        
        The following two snippets are equivalent:
        
        .. highlight:: python
        .. code-block:: python
        
            my_sequence = TestSequence('Parent')
            sub_sequence = my_sequence.sub_sequence('Child')
        
        .. highlight:: python
        .. code-block:: python
        
            my_sequence = TestSequence('Parent')
            sub_sequence = TestSequence('Child')
            my_sequence.append(sub_sequence)
            
        """
        group = TestSequence(name)
        self.append(group)
        return group
    
    def append(self, *phases):
        """ Append normal phases (or sequences) to this test plan.

        Args:
            phases: The phases to append.
        """
        for phase in phases:
            self._test_phases.append(phase)
        
    def _decorate_phase(self, name, array, options=None):
        if not options:
            options = {}
        options['name'] = name
        def _note_fn(fn):
            phase = self._add_phase(fn, array, options)
            return phase
        return _note_fn
    
    def _add_phase(self, fn, array, options):
        phase = ensure_htf_phase(fn)
        phase.options.update(**options)
        array.append(phase)
        return phase
        
    @property
    def phase_group(self):
        # Recursively get phase groups of sub phases if available, else the phase itself.
        _test_phases = [getattr(phase, 'phase_group', phase) for phase in self._test_phases]
        
        return htf.PhaseGroup(
            setup=self._setup_phases,
            main=_test_phases,
            teardown=self._teardown_phases,
            name=self.name
        )
    
class TestPlan(TestSequence):
    """ The core spintop-openhtf interface to create an openhtf sequence.

    :class:`TestPlan` simplifies the creating of openhtf sequences. Instead of declaring
    functions and adding them as an array to openhtf, the test plan allows 
    declarative addition of function using decorators.

    The TestPlan is itself a :class:`TestSequence` and therefore implements all methods
    defined there.

    In OpenHTF, you would do the following:

    .. highlight:: python
    .. code-block:: python

        import openhtf as htf

        @htf.plug(my_plug=...)
        def example_test(test, my_plug):
            time.sleep(.2)

        @htf.plug(my_plug=...)
        def example_test2(test, my_plug):
            time.sleep(.2)

        test = htf.Test(example_test, example_test2)
        test.run() # Run once only

    With the TestPlan, you can do the following (equivalent):
    
    .. highlight:: python
    .. code-block:: python

        from spintop_openhtf import TestPlan

        plan = TestPlan('my-test-name')

        @plan.testcase('Test1')
        @plan.plug(my_plug=...)
        def example_test(test, my_plug):
            time.sleep(.2)

        @plan.testcase('Test2')
        @plan.plug(my_plug=...)
        def example_test2(test, my_plug):
            time.sleep(.2)

        plan.run_once()

    Args:
        name: The name of the test plan. Used to identify the 'type' of the test.
        store_result: Whether results should automatically be stored as JSON in the
            spintop_openhtf site directory.
    """
    DEFAULT_CONF = dict(
        station_server_port='4444', 
        capture_docstring=True
    )

    def __init__(self, name:str='testplan', store_result:bool=True):
        super(TestPlan, self).__init__(name=name)
        
        self._execute_test = None
        self._top_level_component = None
        self.coverage = None
        self.file_provider = TemporaryFileProvider()
        self.callbacks = []
        
        # Array but must contain only one phase.
        # Array is for compatibility with self._decorate_phase function of parent class.
        self._trigger_phases = []
        self._no_trigger = False
        
        if store_result:
            self.add_callbacks(LocalStorageOutput(_local_storage_filename_pattern, indent=4))
            
        self.failure_exceptions = (user_input.SecondaryOptionOccured,)

    @property
    def execute_test(self) -> Callable:
        """Returns a function that takes no arguments and that executes the test described by this test plan."""
        if not self._execute_test:
            self.freeze_test()
        return self._execute_test
    
    @property
    def history_path(self) -> str:
        """The base path where results of the tests defined by this object are stored."""
        path = os.path.join(HISTORY_BASE_PATH, self.name)
        if not os.path.exists(path):
            os.makedirs(path)
        return path
    
    @property
    def is_runnable(self) -> bool:
        """Whether this test plan contains any runnable phases."""
        phases = self._test_phases + self._trigger_phases
        return bool(phases)
    
    @property
    def trigger_phase(self) -> Union[Callable, None]:
        """Returns the defined """
        return self._trigger_phases[0] if self._trigger_phases else None
        
    @property
    def is_test_frozen(self):
        return bool(self._execute_test)
        
    def image_url(self, url:str) -> str:
        """Creates a temporary hosted image based on the specified file path.
        
        Args:
            url: The image path.
        
        Returns:
            The temporary url associated with this image. Can be used in custom forms to show images.
        """
        return self.file_provider.create_url(url)
    
    def trigger(self, name:str, **options):
        """Decorator factory for the trigger phase. 
        
        Similar to :meth:`~TestPlan.testcase`, except that this function will be used as the test trigger.
        
        The test trigger is a special test phase that is executed before test officialy start.
        Once this phase is complete, the test will start. Usually used to configure the test with
        the DUT id for example."""
        if self.trigger_phase:
            raise TestPlanError('There can only be one @trigger function.')
        
        return self._decorate_phase(name, self._trigger_phases, options)

    def no_trigger(self):
        """Removes the need for a trigger and removes the default trigger.
        
        .. highlight:: python
        .. code-block:: python

            plan = TestPlan('my-plan')
            plan.no_trigger()

            # ... 
            
            plan.run() # Will start immediately.

        """
        self._no_trigger = True
    
    def add_callbacks(self, *callbacks):
        """Add custom callbacks to the underlying openhtf test.
        
        Args:
            callbacks:
                The callbacks to add.
        """
        if self.is_test_frozen:
            raise RuntimeError('Cannot add callbacks to the test plan after the test was frozen.')
        self.callbacks += callbacks
    
    def assert_runnable(self):
        if not self.is_runnable:
            # No phases ! Abort now.
            raise RuntimeError('Test is empty, aborting.')
    
    def execute(self):
        """ Execute the configured test using the test_start function as a trigger.
        """
        return self.execute_test()
    
    def run_once(self, launch_browser=True):
        """Shortcut for :meth:`~TestPlan.run` with once=True."""
        return self.run(launch_browser=launch_browser, once=True)
    
    def run(self, launch_browser=True, once=False):
        """Run this test with the OpenHTF frontend. 

        Requires the [server] extra to work. If you do not need the server, you should use
        :meth:`~TestPlan.run_console` which disables completely the GUI server and increases
        execution speed therefore.
        
        Args:
            launch_browser: When True, a browser page will open automatically when this is called.
            once: When False, the test will run in a loop; i.e. when a test ends a new one will start immediately.
        
        """
        conf.load(user_input_enable_console=False, _override=False)
        with self._station_server_context(launch_browser):
            self._run(once)
    
    def run_console(self, once=False):
        """Run this test without a frontend server and enables console input. 

        To enable the server, use :meth:`~TestPlan.run` instead. 

        Args:
            once: When False, the test will run in a loop; i.e. when a test ends a new one will start immediately.
        """
        self._run(once)

    def _run(self, once):
        while True:
            try:
                self.execute()
            except KeyboardInterrupt:
                break
            finally:
                if once:
                    break
    
    def freeze_test(self):
        self._load_default_conf()
        self._execute_test = self._create_execute_test()
        
    @contextmanager
    def _station_server_context(self, launch_browser=True):
        self._load_default_conf() # preload conf so that the server port conf values is populated.
        with station_server.StationServer(self.file_provider) as server:
            self.add_callbacks(server.publish_final_state)
            self.assert_runnable() # Check before launching browser. self.execute also checks when freezing the test.
            
            if launch_browser and conf['station_server_port']:
                webbrowser.open('http://localhost:%s' % conf['station_server_port'])
            
            yield
    
    def _load_default_conf(self):
        conf.load_from_dict(self.DEFAULT_CONF, _override=False)

    def _create_execute_test(self):
        self.assert_runnable()
        
        test = Test(self.phase_group, test_name=self.name, _code_info_after_file=__file__)
        test.configure(failure_exceptions=self.failure_exceptions)
        test.add_output_callbacks(*self.callbacks)
        
        
        trigger_phase = self._build_trigger_phase()
        
        def execute_test():
            return test.execute(test_start=trigger_phase)

        return execute_test
        
    def _build_trigger_phase(self):
        if self.trigger_phase:
            return self.trigger_phase
        elif self._no_trigger:
            return None
        else:
            return create_default_trigger()
    
    def create_plug(self):
        class _SelfReferingPlug(BasePlug):
            def __new__(cls):
                return self
        
        return _SelfReferingPlug
    
    def spintop_plug(self, fn):
        return htf.plugs.plug(spintop=self.create_plug())(fn)
    

def _local_storage_filename_pattern(**test_record):
    folder = '{metadata[test_name]}'.format(**test_record)
    start_time_datetime = datetime.datetime.utcfromtimestamp(test_record['start_time_millis']/1000.0)
    start_time = start_time_datetime.strftime(r"%Y_%m_%d_%H%M%S_%f")
    subfolder = '{dut_id}_{start_time}_{outcome}'.format(start_time=start_time, **test_record)
    return os.path.join(HISTORY_BASE_PATH, folder, subfolder)
    
def create_default_trigger(message='Enter a DUT ID in order to start the test.',
        validator=lambda sn: sn, **state):
    
    @htf.PhaseOptions(timeout_s=None, requires_state=True)
    @htf.plugs.plug(prompts=user_input.UserInput)
    def trigger_phase(state, prompts):
        """Test start trigger that prompts the user for a DUT ID."""
        dut_id = prompts.prompt(message, text_input=True)
        state.test_record.dut_id = validator(dut_id)
        
    return trigger_phase
    
def ensure_htf_phase(fn):
    if not hasattr(fn, 'options'):
        # Not a htf phase, decorate it so it becomes one.
        fn = htf.PhaseOptions()(fn) 
    return fn



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test logic.

Run with (your virtualenv must be activated first):
python all_the_things.py
"""

import time
import os.path

import openhtf as htf
from openhtf import util
from openhtf.util import units
from openhtf.plugs import user_input
from openhtf.output import callbacks
from openhtf.output.callbacks import console_summary
from openhtf.output.callbacks import json_factory

from openhtf.examples import example_plugs


@htf.plug(example=example_plugs.ExamplePlug)
@htf.plug(frontend_aware=example_plugs.ExampleFrontendAwarePlug)
def example_monitor(example, frontend_aware):
  time.sleep(.2)
  frontend_aware.increment()
  return example.increment()


@htf.measures(
    htf.Measurement(
        'widget_type').matches_regex(r'.*Widget$').doc(
            '''This measurement tracks the type of widgets.'''),
    htf.Measurement(
        'widget_color').doc('Color of the widget'),
    htf.Measurement('widget_size').in_range(1, 4).doc('Size of widget'))
@htf.measures('specified_as_args', docstring='Helpful docstring',
              units=units.HERTZ,
              validators=[util.validators.matches_regex('Measurement')])
@htf.plug(example=example_plugs.ExamplePlug)
@htf.plug(prompts=user_input.UserInput)
def hello_world(test, example, prompts):
  """A hello world test phase."""
  test.logger.info('Hello World!')
  test.measurements.widget_type = prompts.prompt(
      'What\'s the widget type? (Hint: try `MyWidget` to PASS)',
      text_input=True)
  if test.measurements.widget_type == 'raise':
    raise Exception()
  test.measurements.widget_color = 'Black'
  test.measurements.widget_size = 3
  test.measurements.specified_as_args = 'Measurement args specified directly'
  test.logger.info('Plug value: %s', example.increment())


# Timeout if this phase takes longer than 10 seconds.
@htf.TestPhase(timeout_s=10)
@htf.measures(
    *(htf.Measurement(
        'level_%s' % i) for i in ['none', 'some', 'all']))
@htf.monitors('monitor_measurement', example_monitor)
def set_measurements(test):
  """Test phase that sets a measurement."""
  test.measurements.level_none = 0
  time.sleep(1)
  test.measurements.level_some = 8
  time.sleep(1)
  test.measurements.level_all = 9
  time.sleep(1)
  level_all = test.get_measurement('level_all')
  assert level_all.value == 9


@htf.measures(
    htf.Measurement('dimensions').with_dimensions(units.HERTZ),
    htf.Measurement('lots_of_dims').with_dimensions(
        units.HERTZ, units.SECOND,
        htf.Dimension(description='my_angle', unit=units.RADIAN)))
def dimensions(test):
  for dim in range(5):
    test.measurements.dimensions[dim] = 1 << dim
  for x, y, z in zip(range(1, 5), range(21, 25), range (101, 105)):
    test.measurements.lots_of_dims[x, y, z] = x + y + z


@htf.measures(
    htf.Measurement('replaced_min_only').in_range('{min}', 5, type=int),
    htf.Measurement('replaced_max_only').in_range(0, '{max}', type=int),
    htf.Measurement('replaced_min_max').in_range('{min}', '{max}', type=int),
)
def measures_with_args(test, min, max):
  test.measurements.replaced_min_only = 1
  test.measurements.replaced_max_only = 1
  test.measurements.replaced_min_max = 1


def attachments(test):
  test.attach('test_attachment', 'This is test attachment data.'.encode('utf-8'))
  test.attach_from_file(
      os.path.join(os.path.dirname(__file__), 'example_attachment.txt'))

  test_attachment = test.get_attachment('test_attachment')
  assert test_attachment.data == b'This is test attachment data.'


@htf.TestPhase(run_if=lambda: False)
def skip_phase(test):
  """Don't run this phase."""


def analysis(test):
  level_all = test.get_measurement('level_all')
  assert level_all.value == 9
  test_attachment = test.get_attachment('test_attachment')
  assert test_attachment.data == b'This is test attachment data.'
  lots_of_dims = test.get_measurement('lots_of_dims')
  assert lots_of_dims.value.value == [
      (1, 21, 101, 123),
      (2, 22, 102, 126),
      (3, 23, 103, 129),
      (4, 24, 104, 132)
  ]
  test.logger.info('Pandas datafram of lots_of_dims \n:%s',
                   lots_of_dims.value.to_dataframe())


def teardown(test):
  test.logger.info('Running teardown')


if __name__ == '__main__':
  test = htf.Test(
      htf.PhaseGroup.with_teardown(teardown)(
          hello_world,
          set_measurements, dimensions, attachments, skip_phase,
          measures_with_args.with_args(min=1, max=4), analysis,
      ),
      # Some metadata fields, these in particular are used by mfg-inspector,
      # but you can include any metadata fields.
      test_name='MyTest', test_description='OpenHTF Example Test',
      test_version='1.0.0')
  test.add_output_callbacks(callbacks.OutputToFile(
      './{dut_id}.{metadata[test_name]}.{start_time_millis}.pickle'))
  test.add_output_callbacks(json_factory.OutputToJSON(
      './{dut_id}.{metadata[test_name]}.{start_time_millis}.json', indent=4))
  test.add_output_callbacks(console_summary.ConsoleSummary())

  # Example of how to output to testrun protobuf format and save to disk then
  # upload.  Replace json_file with your JSON-formatted private key downloaded
  # from Google Developers Console when you created the Service Account you
  # intend to use, or name it 'my_private_key.json'.
  # inspector = (mfg_inspector.MfgInspector
  #    .from_json(json.load(json_file)))
  #    .set_converter(test_runs_converter.test_run_from_test_record))
  # test.add_output_callbacks(
  #     inspector.save_to_disk('./{dut_id}.{start_time_millis}.pb'),
  #     inspector.upload())


  test.execute(test_start=user_input.prompt_for_test_start())



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A simple utility to check whether all previous phases have passed.

In general test execution stops on a raised Exception but will continue if a
phase has a failed measurement.  This is desirable behavior in many cases
because you want to acquire additional information before the test stops.

However, in some cases it's less desirable because downstream long-running
phases will be executed only to FAIL at test completion due to the failing
measurements.  This can have negative implications for tact time.

Checkpoints allow test authors to have their cake and eat it too.  A checkpoint
checks the result of all previous phases and will stop test execution if a
previous phase has failed.
"""

from openhtf.core import phase_descriptor
from openhtf.core import test_record

def checkpoint(checkpoint_name=None):
  name = checkpoint_name if checkpoint_name else 'Checkpoint'

  @phase_descriptor.PhaseOptions(name=name)
  def _checkpoint(test_run):
    failed_phases = []
    for phase_record in test_run.test_record.phases:
      if phase_record.outcome == test_record.PhaseOutcome.FAIL:
        failed_phases.append(phase_record.name)

    if failed_phases:
      test_run.logger.error('Stopping execution because phases failed: %s',
                            failed_phases)
      return phase_descriptor.PhaseResult.STOP

  return _checkpoint



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Stub USB handle implementation for testing."""

import binascii
import string

from openhtf.plugs.usb import usb_handle


class StubUsbHandle(usb_handle.UsbHandle):
  """Stub handle used for testing."""
  PRINTABLE_DATA = set(string.printable) - set(string.whitespace)

  def __init__(self, ignore_writes=False):
    super(StubUsbHandle, self).__init__('StubSerial', 'StubHandle',
                                        default_timeout_ms=0)
    self.expected_write_data = None if ignore_writes else []
    self.expected_read_data = []
    self.closed = False

  @classmethod
  def _dotify(cls, data):
    """Add dots."""
    return ''.join(char if char in cls.PRINTABLE_DATA else '.' for char in data)

  def write(self, data, dummy=None):
    """Stub Write method."""
    assert not self.closed
    if self.expected_write_data is None:
      return

    expected_data = self.expected_write_data.pop(0)
    if expected_data != data:
      raise ValueError('Expected %s, got %s (%s)' % (
          self._dotify(expected_data), binascii.hexlify(data),
          self._dotify(data)))

  def read(self, length, dummy=None):
    """Stub Read method."""
    assert not self.closed
    data = self.expected_read_data.pop(0)
    if length < len(data):
      raise ValueError(
          'Overflow packet length. Read %d bytes, got %d bytes: %s',
          length, len(data), self._dotify(data))
    return data

  def close(self):
    """Stub Close method."""
    self.closed = True

  def is_closed(self):
    """Stub is_closed method."""
    return self.closed

  def expect_write(self, data):
    """Stub expect_write method."""
    assert self.expected_write_data is not None, \
        'expect_write would be ignored!'
    self.expected_write_data.append(data)

  def expect_read(self, data):
    """Stub expect_read method."""
    self.expected_read_data.append(data)



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import time
from openhtf.plugs.usb import local_usb

class EtherSync(object):
  """EtherSync object for the access of usb device connected to
     Cambrionix unit."""

  port_map = {
      '1':'112',
      '2':'111',
      '3':'114',
      '4':'113',
      '5':'212',
      '6':'211',
      '7':'214',
      '8':'213',
  }

  def __init__(self, mac_addr):
    """Construct a EtherSync object.

      Args:
        mac_addr: mac address of the Cambrionix unit for EtherSync.
    """
    addr_info = mac_addr.lower().split(':')
    if len(addr_info) < 6:
      raise ValueError('Invalid mac address')

    addr_info[2] = 'EtherSync'
    self._addr = ''.join(addr_info[2:])

  def get_usb_serial(self, port_num):
    """Get the device serial number

    Args:
      port_num: port number on the Cambrionix unit

    Return:
      usb device serial number
    """

    port = self.port_map[str(port_num)]
    arg = ''.join(['DEVICE INFO,', self._addr, '.', port])
    cmd = (['esuit64', '-t', arg])
    info = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    serial = None
    if "SERIAL" in info:
      serial_info = info.split('SERIAL:')[1]
      serial = serial_info.split('\n')[0].strip()
      use_info = info.split('BY')[1].split(' ')[1]
      if use_info == 'NO':
        cmd = (['esuit64', '-t', 'AUTO USE ALL'])
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        time.sleep(50.0/1000.0)
    else:
      raise ValueError('No USB device detected')
    return serial

  def open_usb_handle(self, port_num):
    """open usb port

    Args:
      port_num: port number on the Cambrionix unit

    Return:
      usb handle
    """
    serial = self.get_usb_serial(port_num)
    return local_usb.LibUsbHandle.open(serial_number=serial)



import pytest
from mock import MagicMock
import paramiko

from spintop_openhtf.plugs.ssh import SSHInterface


ADDR = 'x'
USER = 'y'
PASS = 'z'
TIMEOUT = 23

def test_ssh_client_created():
    ssh_interface = SSHInterface(ADDR, USER, PASS, create_timeout=TIMEOUT)
    ssh_client_mock = MagicMock(paramiko.SSHClient)
    ssh_interface.open(ssh_client_mock)

    assert isinstance(ssh_client_mock.set_missing_host_key_policy.call_args[0][0], paramiko.AutoAddPolicy)
    ssh_client_mock.connect.assert_called_with(ADDR, port=22, username=USER, password=PASS, timeout=TIMEOUT)

def test_ssh_client_close_works():
    ssh_interface = SSHInterface(ADDR, USER, PASS, create_timeout=TIMEOUT)
    ssh_client_mock = MagicMock(paramiko.SSHClient)
    ssh_interface.open(ssh_client_mock)
    ssh_interface.close()

    ssh_client_mock.close.assert_called_once()






# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020

"""Monitors provide a mechanism for periodically collecting data and
automatically persisting values in a measurement.

Monitors are implemented similar to phase functions - they are decorated
with plugs.plug() to pass plugs in.  The return value of a monitor
function, however, will be used to append a value to a measurement.

Monitors by default poll at a rate of 1 second between invocations of
the monitor function.  The poll interval (given in milliseconds) determines the
approximate frequency at which values will be sampled.  A sample is considered
to have been taken at the time when the monitor function *returns*, not when
it is called.

The approximate average duration of calls to the monitor function is taken into
account, so that samples are obtained on as close to interval_ms boundaries as
can be.  A poll interval of 0 will cause the monitor function to be called in a
tight loop with no delays.

Example:

@plugs.plug(current_meter=current_meter.CurrentMeter)
def CurrentMonitor(test, current_meter):
  return current_meter.GetReading()

@monitors.monitors('current_draw', CurrentMonitor, units=units.AMPERE)
def MyPhase(test):
  # Do some stuff for a while...

# MyPhase will have a dimensioned measurement on it, with units of 'AMPERE' and
# a single dimension of 'MILLISECONDS', and will have values for roughly every
# second while MyPhase was executing.
"""

import functools
import inspect
import time

import openhtf
from openhtf import plugs
from openhtf.core import measurements
from openhtf.util import threads
from openhtf.util import units as uom
from openhtf.util import functions


class _MonitorThread(threads.KillableThread):

  daemon = True

  def __init__(self, measurement_name, monitor_desc, extra_kwargs, test_state,
               interval_ms):
    super(_MonitorThread, self).__init__(
        name='%s_MonitorThread' % measurement_name)
    self.measurement_name = measurement_name
    self.monitor_desc = monitor_desc
    self.test_state = test_state
    self.interval_ms = interval_ms
    self.extra_kwargs = extra_kwargs

  def get_value(self):
    arg_info = functions.getargspec(self.monitor_desc.func)
    if arg_info.keywords:
      # Monitor phase takes **kwargs, so just pass everything in.
      kwargs = self.extra_kwargs
    else:
      # Only pass in args that the monitor phase takes.
      kwargs = {arg: val for arg, val in self.extra_kwargs
                if arg in arg_info.args}
    return self.monitor_desc.with_args(**kwargs)(self.test_state)

  def _thread_proc(self):
    measurement = getattr(self.test_state.test_api.measurements,
                          self.measurement_name)
    start_time = time.time()

    # Special case tight-loop monitoring.
    if not self.interval_ms:
      while True:
        measurement[(time.time() - start_time) * 1000] = self.get_value()

    # Helper to take sample, return sample number and sample duration.
    def _take_sample():
      pre_time, value, post_time = time.time(), self.get_value(), time.time()
      measurement[(post_time - start_time) * 1000] = value
      return (int((post_time - start_time) * 1000 / self.interval_ms),
              (post_time - pre_time) * 1000)

    # Track the last sample number, and an approximation of the mean time
    # it takes to sample (so we can account for it in how long we sleep).
    last_sample, mean_sample_ms = _take_sample()
    while True:
      # Find what sample number (float) we would be on if we sampled now.
      current_time = time.time()
      new_sample = ((((current_time - start_time) * 1000) + mean_sample_ms) /
                    self.interval_ms)
      if new_sample < last_sample + 1:
        time.sleep(start_time - current_time +
                   ((last_sample + 1) * self.interval_ms / 1000.0) -
                   (mean_sample_ms / 1000.0))
        continue
      elif new_sample > last_sample + 2:
        self.test_state.state_logger.warning(
            'Monitor for "%s" skipping %s sample(s).', self.measurement_name,
            new_sample - last_sample - 1)
      last_sample, cur_sample_ms = _take_sample()
      # Approximate 10-element sliding window average.
      mean_sample_ms = ((9 * mean_sample_ms) + cur_sample_ms) / 10.0


def monitors(measurement_name, monitor_func, units=None, poll_interval_ms=1000):
  monitor_desc = openhtf.PhaseDescriptor.wrap_or_copy(monitor_func)
  def wrapper(phase_func):
    phase_desc = openhtf.PhaseDescriptor.wrap_or_copy(phase_func)

    # Re-key this dict so we don't have to worry about collisions with
    # plug.plug() decorators on the phase function.  Since we aren't
    # updating kwargs here, we don't have to worry about collisions with
    # kwarg names.
    monitor_plugs = {('_' * idx) + measurement_name + '_monitor': plug.cls for
                     idx, plug in enumerate(monitor_desc.plugs, start=1)}

    @openhtf.PhaseOptions(requires_state=True)
    @plugs.plug(update_kwargs=False, **monitor_plugs)
    @measurements.measures(
        measurements.Measurement(measurement_name).with_units(
            units).with_dimensions(uom.MILLISECOND))
    @functools.wraps(phase_desc.func)
    def monitored_phase_func(test_state, *args, **kwargs):
      # Start monitor thread, it will run monitor_desc periodically.
      monitor_thread = _MonitorThread(
          measurement_name, monitor_desc, phase_desc.extra_kwargs, test_state,
          poll_interval_ms)
      monitor_thread.start()
      try:
        return phase_desc(test_state, *args, **kwargs)
      finally:
        monitor_thread.kill()
        monitor_thread.join()
    return monitored_phase_func
  return wrapper



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for utility functions that manipulate or compare data.

We use a few special data formats internally, these utility functions make it a
little easier to work with them.
"""

import collections
import difflib
import itertools
import logging
import math
import numbers
import pprint
import struct
import sys

from mutablerecords import records
from past.builtins import long
from past.builtins import unicode

from enum import Enum
import six

# Used by convert_to_base_types().
PASSTHROUGH_TYPES = {bool, bytes, int, long, type(None), unicode}


def pprint_diff(first, second, first_name='first', second_name='second'):
  """Compare the pprint representation of two objects and yield diff lines."""
  return difflib.unified_diff(
      pprint.pformat(first).splitlines(),
      pprint.pformat(second).splitlines(),
      fromfile=first_name, tofile=second_name, lineterm='')


def equals_log_diff(expected, actual, level=logging.ERROR):
  """Compare two string blobs, error log diff if they don't match."""
  if expected == actual:
    return True

  # Output the diff first.
  logging.log(level, '***** Data mismatch: *****')
  for line in difflib.unified_diff(
      expected.splitlines(), actual.splitlines(),
      fromfile='expected', tofile='actual', lineterm=''):
    logging.log(level, line)
  logging.log(level, '^^^^^  Data diff  ^^^^^')


def assert_records_equal_nonvolatile(first, second, volatile_fields, indent=0):
  """Compare two test_record tuples, ignoring any volatile fields.

  'Volatile' fields include any fields that are expected to differ between
  successive runs of the same test, mainly timestamps.  All other fields
  are recursively compared.
  """
  if isinstance(first, dict) and isinstance(second, dict):
    if set(first) != set(second):
      logging.error('%sMismatching keys:', ' ' * indent)
      logging.error('%s  %s', ' ' * indent, list(first.keys()))
      logging.error('%s  %s', ' ' * indent, list(second.keys()))
      assert set(first) == set(second)
    for key in first:
      if key in volatile_fields:
        continue
      try:
        assert_records_equal_nonvolatile(first[key], second[key],
                                         volatile_fields, indent + 2)
      except AssertionError:
        logging.error('%sKey: %s ^', ' ' * indent, key)
        raise
  elif hasattr(first, '_asdict') and hasattr(second, '_asdict'):
    # Compare namedtuples as dicts so we get more useful output.
    assert_records_equal_nonvolatile(first._asdict(), second._asdict(),
                                     volatile_fields, indent)
  elif hasattr(first, '__iter__') and hasattr(second, '__iter__'):
    for idx, (fir, sec) in enumerate(itertools.izip(first, second)):
      try:
        assert_records_equal_nonvolatile(fir, sec, volatile_fields, indent + 2)
      except AssertionError:
        logging.error('%sIndex: %s ^', ' ' * indent, idx)
        raise
  elif (isinstance(first, records.RecordClass) and
        isinstance(second, records.RecordClass)):
    assert_records_equal_nonvolatile(
        {slot: getattr(first, slot) for slot in first.__slots__},
        {slot: getattr(second, slot) for slot in second.__slots__},
        volatile_fields, indent)
  elif first != second:
    logging.error('%sRaw: "%s" != "%s"', ' ' * indent, first, second)
    assert first == second


def convert_to_base_types(obj, ignore_keys=tuple(), tuple_type=tuple,
                          json_safe=True):
  """Recursively convert objects into base types.

  This is used to convert some special types of objects used internally into
  base types for more friendly output via mechanisms such as JSON.  It is used
  for sending internal objects via the network and outputting test records.
  Specifically, the conversions that are performed:

    - If an object has an as_base_types() method, immediately return the result
      without any recursion; this can be used with caching in the object to
      prevent unnecessary conversions.
    - If an object has an _asdict() method, use that to convert it to a dict and
      recursively converting its contents.
    - mutablerecords Record instances are converted to dicts that map
      attribute name to value.  Optional attributes with a value of None are
      skipped.
    - Enum instances are converted to strings via their .name attribute.
    - Real and integral numbers are converted to built-in types.
    - Byte and unicode strings are left alone (instances of six.string_types).
    - Other non-None values are converted to strings via str().

  The return value contains only the Python built-in types: dict, list, tuple,
  str, unicode, int, float, long, bool, and NoneType (unless tuple_type is set
  to something else).  If tuples should be converted to lists (e.g. for an
  encoding that does not differentiate between the two), pass 'tuple_type=list'
  as an argument.

  If `json_safe` is True, then the float 'inf', '-inf', and 'nan' values will be
  converted to strings. This ensures that the returned dictionary can be passed
  to json.dumps to create valid JSON. Otherwise, json.dumps may return values
  such as NaN which are not valid JSON.
  """
  # Because it's *really* annoying to pass a single string accidentally.
  assert not isinstance(ignore_keys, six.string_types), 'Pass a real iterable!'

  if hasattr(obj, 'as_base_types'):
    return obj.as_base_types()
  if hasattr(obj, '_asdict'):
    obj = obj._asdict()
  elif isinstance(obj, records.RecordClass):
    obj = {attr: getattr(obj, attr)
           for attr in type(obj).all_attribute_names
           if (getattr(obj, attr, None) is not None or
               attr in type(obj).required_attributes)}
  elif isinstance(obj, Enum):
    obj = obj.name

  if type(obj) in PASSTHROUGH_TYPES:
    return obj

  # Recursively convert values in dicts, lists, and tuples.
  if isinstance(obj, dict):
    return {convert_to_base_types(k, ignore_keys, tuple_type):
               convert_to_base_types(v, ignore_keys, tuple_type)
            for k, v in six.iteritems(obj) if k not in ignore_keys}
  elif isinstance(obj, list):
    return [convert_to_base_types(val, ignore_keys, tuple_type, json_safe)
            for val in obj]
  elif isinstance(obj, tuple):
    return tuple_type(
        convert_to_base_types(value, ignore_keys, tuple_type, json_safe)
        for value in obj)

  # Convert numeric types (e.g. numpy ints and floats) into built-in types.
  elif isinstance(obj, numbers.Integral):
    return long(obj)
  elif isinstance(obj, numbers.Real):
    as_float = float(obj)
    if json_safe and (math.isinf(as_float) or math.isnan(as_float)):
      return str(as_float)
    return as_float

  # Convert all other types to strings.
  try:
    return str(obj)
  except:
    logging.warning('Problem casting object of type %s to str.', type(obj))
    raise


def total_size(obj):
  """Returns the approximate total memory footprint an object."""
  seen = set()
  def sizeof(current_obj):
    try:
      return _sizeof(current_obj)
    except Exception:  # pylint: disable=broad-except
      # Not sure what just happened, but let's assume it's a reference.
      return struct.calcsize('P')

  def _sizeof(current_obj):
    """Do a depth-first acyclic traversal of all reachable objects."""
    if id(current_obj) in seen:
      # A rough approximation of the size cost of an additional reference.
      return struct.calcsize('P')
    seen.add(id(current_obj))
    size = sys.getsizeof(current_obj)

    if isinstance(current_obj, dict):
      size += sum(map(sizeof, itertools.chain.from_iterable(
          six.iteritems(current_obj))))
    elif (isinstance(current_obj, collections.Iterable) and
          not isinstance(current_obj, six.string_types)):
      size += sum(sizeof(item) for item in current_obj)
    elif isinstance(current_obj, records.RecordClass):
      size += sum(sizeof(getattr(current_obj, attr))
                  for attr in current_obj.__slots__)
    return size

  return sizeof(obj)



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020


"""OpenHTF module responsible for managing records of tests."""


import collections
import hashlib
import inspect
import logging
import os

from enum import Enum

import mutablerecords

from openhtf import util
from openhtf.util import data
from openhtf.util import logs

import six

_LOG = logging.getLogger(__name__)


class InvalidMeasurementDimensions(Exception):
  """Raised when a measurement is taken with the wrong number of dimensions."""


OutcomeDetails = collections.namedtuple(
    'OutcomeDetails', 'code description')
Outcome = Enum('Outcome', ['PASS', 'FAIL', 'ERROR', 'TIMEOUT', 'ABORTED'])  # pylint: disable=invalid-name
# LogRecord is in openhtf.util.logs.LogRecord.


class Attachment(collections.namedtuple('Attachment', 'data mimetype')):
  """Encapsulate attachment data and guessed MIME type."""

  @property
  def sha1(self):
    return hashlib.sha1(six.ensure_binary(self.data)).hexdigest()

  def _asdict(self):
    # Don't include the attachment data when converting to dict.
    return {
        'mimetype': self.mimetype,
        'sha1': self.sha1,
    }

class TestRecord(  # pylint: disable=no-init
    mutablerecords.Record(
        'TestRecord', ['dut_id', 'station_id'],
        {'start_time_millis': int,
         'end_time_millis': None,
         'outcome': None,
         'outcome_details': list,
         'code_info': None,
         'metadata': dict,
         'phases': list,
         'log_records': list,
         '_cached_record': dict,
         '_cached_phases': list,
         '_cached_log_records': list,
         '_cached_config_from_metadata': dict,
        })):
  """The record of a single run of a test."""

  def __init__(self, *args, **kwargs):
    super(TestRecord, self).__init__(*args, **kwargs)
    # Cache data that does not change during execution.
    # Cache the metadata config so it does not recursively copied over and over
    # again.
    self._cached_config_from_metadata = self.metadata.get('config')
    self._cached_record = {
        'station_id': data.convert_to_base_types(self.station_id),
        'code_info': data.convert_to_base_types(self.code_info),
    }

  def add_outcome_details(self, code, description=''):
    """Adds a code with optional description to this record's outcome_details.

    Args:
      code: A code name or number.
      description: A string providing more details about the outcome code.
    """
    self.outcome_details.append(OutcomeDetails(code, description))

  def add_phase_record(self, phase_record):
    self.phases.append(phase_record)
    self._cached_phases.append(phase_record.as_base_types())

  def add_log_record(self, log_record):
    self.log_records.append(log_record)
    self._cached_log_records.append(log_record._asdict())

  def as_base_types(self):
    """Convert to a dict representation composed exclusively of base types."""
    metadata = data.convert_to_base_types(self.metadata,
                                          ignore_keys=('config',))
    metadata['config'] = self._cached_config_from_metadata
    ret = {
        'dut_id': data.convert_to_base_types(self.dut_id),
        'start_time_millis': self.start_time_millis,
        'end_time_millis': self.end_time_millis,
        'outcome': data.convert_to_base_types(self.outcome),
        'outcome_details': data.convert_to_base_types(self.outcome_details),
        'metadata': metadata,
        'phases': self._cached_phases,
        'log_records': self._cached_log_records,
    }
    ret.update(self._cached_record)
    return ret
  
  def is_started(self):
    return bool(self.start_time_millis)
  
# PhaseResult enumerations are converted to these outcomes by the PhaseState.
PhaseOutcome = Enum(  # pylint: disable=invalid-name
    'PhaseOutcome', [
        'PASS',  # CONTINUE with allowed measurement outcomes.
        'FAIL',  # CONTINUE with failed measurements or FAIL_AND_CONTINUE.
        'SKIP',  # SKIP or REPEAT when under the phase's repeat limit.
        'ERROR',  # Any terminal result.
    ])


class PhaseRecord(  # pylint: disable=no-init
    mutablerecords.Record(
        'PhaseRecord', ['descriptor_id', 'name', 'codeinfo'],
        {'measurements': None, 'options': None,
         'start_time_millis': int, 'end_time_millis': None,
         'attachments': dict, 'result': None, 'outcome': None})):
  """The record of a single run of a phase.

  Measurement metadata (declarations) and values are stored in separate
  dictionaries, each of which map measurement name to the respective object.  In
  the case of the measurements field, those objects are measurements.Measurement
  instances.  The 'value' attribute of each of those instances is an instance of
  measurments.MeasuredValue, which contains either a single value, or a list of
  values in the case of a dimensioned measurement.

  See measurements.Record.GetValues() for more information.

  The 'result' attribute contains a phase_executor.PhaseExecutionOutcome
  instance, which wraps the openhtf.PhaseResult returned by the phase or an
  error condition that terminated the phase.

  The 'outcome' attribute is a PhaseOutcome, which caches the pass/fail outcome
  of the phase's measurements or indicates that the verification was skipped.
  """

  @classmethod
  def from_descriptor(cls, phase_desc):
    return cls(id(phase_desc), phase_desc.name, phase_desc.code_info)

  def as_base_types(self):
    """Convert to a dict representation composed exclusively of base types."""
    base_types_dict = {
        k: data.convert_to_base_types(getattr(self, k))
        for k in self.optional_attributes
    }
    base_types_dict.update(
        descriptor_id=self.descriptor_id,
        name=self.name,
        codeinfo=data.convert_to_base_types(self.codeinfo),
    )
    return base_types_dict

  def record_start_time(self):
    """Record the phase start time and return it."""
    self.start_time_millis = util.time_millis()
    return self.start_time_millis

  def finalize_phase(self, options):
    self.end_time_millis = util.time_millis()
    self.options = options


def _get_source_safely(obj):
  try:
    return inspect.getsource(obj)
  except Exception:  # pylint: disable=broad-except
    logs.log_once(
        _LOG.warning,
        'Unable to load source code for %s. Only logging this once.', obj)
    return ''


class CodeInfo(mutablerecords.HashableRecord(
    'CodeInfo', ['name', 'docstring', 'sourcecode'])):
  """Information regarding the running tester code."""

  @classmethod
  def for_module_from_stack(cls, levels_up=1, after_file=None, with_source=False):
    # levels_up is how many frames up to go:
    #  0: This function (useless).
    #  1: The function calling this (likely).
    #  2+: The function calling 'you' (likely in the framework).
    stack = inspect.stack(context=0)
    if after_file:
      found_file = False
      for stackframe in stack:
        frame, filename = stackframe[:2]
        if filename == after_file:
          found_file = True
        elif found_file:
          # we skipped the file. All good !
          break
        
    else:
      frame, filename = stack[levels_up][:2]
    
    module = inspect.getmodule(frame)
    if with_source:
      source = _get_source_safely(frame)
    else:
      source = ''
    return cls(os.path.basename(filename), inspect.getdoc(module), source)

  @classmethod
  def for_function(cls, func, with_source=False):
    if with_source:
      source = _get_source_safely(func)
    else:
      source = ''
    return cls(func.__name__, inspect.getdoc(func), source)

  @classmethod
  def uncaptured(cls):
    return cls('', None, '')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openhtf/output/proto/guzzle.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='openhtf/output/proto/guzzle.proto',
  package='openhtf',
  syntax='proto2',
  serialized_pb=_b('\n!openhtf/output/proto/guzzle.proto\x12\x07openhtf\"N\n\x0fTestRunEnvelope\x12*\n\x0cpayload_type\x18\x02 \x01(\x0e\x32\x14.openhtf.PayloadType\x12\x0f\n\x07payload\x18\x03 \x01(\x0c*S\n\x0bPayloadType\x12\x17\n\x13\x43OMPRESSED_TEST_RUN\x10\x01\x12\x13\n\x0f\x43OMPRESSED_VARZ\x10\x02\x12\x16\n\x12\x43OMPRESSED_LOGFILE\x10\x03')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_PAYLOADTYPE = _descriptor.EnumDescriptor(
  name='PayloadType',
  full_name='openhtf.PayloadType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COMPRESSED_TEST_RUN', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COMPRESSED_VARZ', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COMPRESSED_LOGFILE', index=2, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=126,
  serialized_end=209,
)
_sym_db.RegisterEnumDescriptor(_PAYLOADTYPE)

PayloadType = enum_type_wrapper.EnumTypeWrapper(_PAYLOADTYPE)
COMPRESSED_TEST_RUN = 1
COMPRESSED_VARZ = 2
COMPRESSED_LOGFILE = 3



_TESTRUNENVELOPE = _descriptor.Descriptor(
  name='TestRunEnvelope',
  full_name='openhtf.TestRunEnvelope',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='payload_type', full_name='openhtf.TestRunEnvelope.payload_type', index=0,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='payload', full_name='openhtf.TestRunEnvelope.payload', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=46,
  serialized_end=124,
)

_TESTRUNENVELOPE.fields_by_name['payload_type'].enum_type = _PAYLOADTYPE
DESCRIPTOR.message_types_by_name['TestRunEnvelope'] = _TESTRUNENVELOPE
DESCRIPTOR.enum_types_by_name['PayloadType'] = _PAYLOADTYPE

TestRunEnvelope = _reflection.GeneratedProtocolMessageType('TestRunEnvelope', (_message.Message,), dict(
  DESCRIPTOR = _TESTRUNENVELOPE,
  __module__ = 'openhtf.output.proto.guzzle_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.TestRunEnvelope)
  ))
_sym_db.RegisterMessage(TestRunEnvelope)


# @@protoc_insertion_point(module_scope)



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface to OpenHTF configuration files.

As a matter of convention, OpenHTF configuration files should contain values
which are specific to an individual station (not station type).  This is
intended to provide a means to decouple deployment of test code from
station-specific configuration or calibration.

Examples of the types of values commonly found in the configuration are
physical port names, IP addresses, calibrated light/sound levels, etc.
Configuration values should not be used to determine test flow, or to control
debug output.

Config keys must be declared as in the following example, where default_value
and description are optional:

  from openhtf.util import conf

  conf.declare('antimatter_intermix_constant',
               default_value=3.14159,
               description='Intermix constant calibrated for our warp core.')

Declared keys can be accessed directly as attributes of the conf module.  To
avoid naming conflicts, configuration keys must begin with a lowercase letter.
They may also be accessed by treating the conf module as a dictionary, but this
method is discouraged and should only be used in favor of getattr().

  from openhtf.util import conf

  warp_core.SetIntermixConstant(conf.antimatter_intermix_constant)

  # An example of when you might use dict-like access.
  for idx in range(5):
    warp_core.SetDilithiumRatio(idx, conf['dilthium_ratio_%s' % idx])

Another common mechanism for obtaining configuration values is to use the
conf.inject_positional_args decorator:

  from openhtf.util import conf

  @conf.inject_positional_args
  def ModifyThePhaseVariance(antimatter_intermix_constant, phase_variance):
    return antimatter_intermix_constant * phase_variance

  # antimatter_intermix_constant will be taken from the configuration value.
  x = ModifyThePhaseVariance(phase_variance=2.71828)

Decorating a function with conf.inject_positional_args forces all other
arguments to be passed by keyword in order to avoid ambiguity in the values of
positional args.  Values passed via keyword that also exist in the config will
override config values and log a warning message.  Keyword args in the function
declaration will not be overridden (because it would be ambiguous which default
to use), and any overlap in keyword arg names and config keys will result in a
warning message.

If the configuration key is declared but no default_value is provided and no
value has been loaded, then no value will be passed, and a TypeError will be
raised unless the value is passed via keyword.  Essentially, if `keyword_arg in
conf` evaluates to True, then that keyword arg will be provided from the
configuration unless overridden in the kwargs passed to the function.  Otherwise
keyword_arg must be passed via kwargs at function invocation time.

The conf module supports 'in' checks, where `key in conf` will evaluate to True
if conf[key] would successfully provide a value.  That is, if either a value
has been loaded or a default_value was declared.

Configuration values may be loaded directly or from a yaml or json file.  If no
configuration is loaded, default values will still be accessible.  Loading a
configuration always overrides default values, but only overrides previously
loaded values if _override=True (default) for the load* method used.  Some
examples of how to load a configuration:

  from openhtf.util import conf

  conf.declare('antimatter_intermix_constant')
  conf.declare('phase_variance')

  conf.load(antimatter_intermix_constant=3.14,
            phase_variance=2.718)
  conf.load_from_dict({
      'antimatter_intermix_constant': 3.14,
      'phase_variance': 2.718,
  })
  conf.load_from_file('config.json')
  conf.load_from_file('config.yaml')

Note that any of the load* methods here accept an _override keyword argument
that defaults to True, but may be set False to prevent overriding previously
loaded values.  Regardless of whether _override is True or False, a message
will be logged indicating how the duplicate value was handled.

conf.load_from_file() attempts to parse the filename given as JSON and as YAML,
if neither succeeds, an exception will be raised.  In either case, the value
parsed must be a dictionary mapping configuration key to value.  Complex
configuration values are discouraged; they should be kept to single values or
lists of values when possible.

Lastly, configuration values may also be provided via the --config-value flag,
but this is discouraged, and should only be used for debugging purposes.

Configuration values loaded via commandline flags, either --config-file or
--config-value, are not checked against Declarations.  This allows for using
configuration files that are supersets of required configuration.  Declarations
are *always* checked upon configuration value access, however, so you still
must declare any keys you wish to use.

Loaded configuration values may be purged via the reset() method, but this
should only be used for testing purposes.  This will reset the configuration
state to what it was before any load* methods were called (defaults loaded
and flag values used, either directly or from --config-file).

A recommended alternative to using reset() is the @save_and_restore decorator,
which allows you to decorate a function or method so that during execution
of the decorated callable, configuration values are altered (and restored
after execution of that callable).  For example:

  conf.load(foo='foo')

  @conf.save_and_restore(foo='bar')
  def do_stuff():
    print 'foo has value: ', conf.foo

  print 'foo before call: ', conf.foo
  do_stuff()
  print 'foo after call: ', conf.foo

This example prints:

  foo before call: foo
  foo has value: bar
  foo after call: foo

This is useful primarily for unittest methods (see util/test.py for specific
examples of unittest usages).  Note that config overrides may be specified at
decoration time, but do not have to be:

  @conf.save_and_restore
  def do_stuff():
    conf.foo = 'bar'

This is also valid.  The entire configuration is restored to the state it had
upon execution of the decorated callable, regardless of which keys are updated
in the decorator or in the decorated callable.
"""

import argparse
import functools
import inspect
import logging
import sys
import threading
import yaml

import mutablerecords
import six


from . import argv
from . import threads
from . import functions

# If provided, --config-file will cause the given file to be load()ed when the
# conf module is initially imported.
ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument(
    '--config-file', type=argparse.FileType('r'),
    help='File from which to load configuration values.')

ARG_PARSER.add_argument(
    '--config-value', action='append', default=[],
    help='Allows specifying a configuration key=value on the command line. '
    'The format should be --config-value=key=value. This value will override '
    'any loaded value, and will be a string.')


class Configuration(object):  # pylint: disable=too-many-instance-attributes
  """A singleton class to replace the 'conf' module.

  This class provides the configuration interface described in the module
  docstring.  All attribuets/methods must not begin with a lowercase letter so
  as to avoid naming conflicts with configuration keys.
  """

  class ConfigurationInvalidError(Exception):
    """Indicates the configuration format was invalid or couldn't be read."""

  class KeyAlreadyDeclaredError(Exception):
    """Indicates that a configuration key was already declared."""

  class UndeclaredKeyError(Exception):
    """Indicates that a key was required but not predeclared."""

  class InvalidKeyError(Exception):
    """Raised when an invalid key is declared or accessed."""

  class UnsetKeyError(Exception):
    """Raised when a key value is requested but we have no value for it."""

  # pylint: disable=invalid-name,bad-super-call
  class Declaration(mutablerecords.Record(
      'Declaration', ['name'], {
          'description': None, 'default_value': None, 'has_default': False})):
    """Record type encapsulating information about a config declaration."""
    def __init__(self, *args, **kwargs):
      super(type(self), self).__init__(*args, **kwargs)
      # Track this separately to allow for None as a default value, override
      # any value that was passed in explicitly - don't do that.
      self.has_default = 'default_value' in kwargs
  # pylint: enable=invalid-name,bad-super-call

  __slots__ = ('_logger', '_lock', '_modules', '_declarations',
               '_flag_values', '_flags', '_loaded_values', 'ARG_PARSER',
               '__name__')

  def __init__(self, logger, lock, parser, **kwargs):
    """Initializes the configuration state.

    We have to pull everything we need from global scope into here because we
    will be swapping out the module with this instance and will lose any global
    references.

    Args:
      logger: Logger to use for logging messages within this class.
      lock: Threading.lock to use for locking access to config values.
      **kwargs: Modules we need to access within this class.
    """
    self._logger = logger
    self._lock = lock
    self._modules = kwargs
    self._declarations = {}
    self.ARG_PARSER = parser

    # Parse just the flags we care about, since this happens at import time.
    self._flags, _ = parser.parse_known_args()
    self._flag_values = {}

    # Populate flag_values from flags now.
    self.load_flag_values()

    # Initialize self._loaded_values and load from --config-file if it's set.
    self.reset()

  def load_flag_values(self, flags=None):
    """Load flag values given from command line flags.

    Args:
      flags: An argparse Namespace containing the command line flags.
    """
    if flags is None:
      flags = self._flags
    for keyval in flags.config_value:
      k, v = keyval.split('=', 1)
      v = self._modules['yaml'].load(v, Loader=self._modules['yaml'].SafeLoader) if isinstance(v, str) else v

      # Force any command line keys and values that are bytes to unicode.
      k = k.decode() if isinstance(k, bytes) else k
      v = v.decode() if isinstance(v, bytes) else v

      self._flag_values.setdefault(k, v)

  @staticmethod
  def _is_valid_key(key):
    """Return True if key is a valid configuration key."""
    return key and key[0].islower()

  def __setattr__(self, attr, value):
    """Provide a useful error when attempting to set a value via setattr()."""
    if self._is_valid_key(attr):
      raise AttributeError("Can't set conf values by attribute, use load()")
    # __slots__ is defined above, so this will raise an AttributeError if the
    # attribute isn't one we expect; this limits the number of ways to abuse the
    # conf module singleton instance.  Also note that we can't use super()
    # normally here because of the sys.modules swap (Configuration is no longer
    # defined, and evaluates to None if used here).
    # pylint: disable=bad-super-call
    super(type(self), self).__setattr__(attr, value)

  # Don't use synchronized on this one, because __getitem__ handles it.
  def __getattr__(self, attr):  # pylint: disable=invalid-name
    """Get a config value via attribute access."""
    if self._is_valid_key(attr):
      return self[attr]
    # Config keys all begin with a lowercase letter, so treat this normally.
    raise AttributeError("'%s' object has no attribute '%s'" %
                         (type(self).__name__, attr))

  @threads.synchronized
  def __getitem__(self, item):  # pylint: disable=invalid-name
    """Get a config value via item access.

    Order of precedence is:
      - Value provided via --config-value flag.
      - Value loaded via load*() methods.
      - Default value as declared with conf.declare()

    Args:
      item: Config key name to get.
    """
    if item not in self._declarations:
      raise self.UndeclaredKeyError('Configuration key not declared', item)

    if item in self._flag_values:
      if item in self._loaded_values:
        self._logger.warning(
            'Overriding loaded value for %s (%s) with flag value: %s',
            item, self._loaded_values[item], self._flag_values[item])
      return self._flag_values[item]
    if item in self._loaded_values:
      return self._loaded_values[item]
    if self._declarations[item].has_default:
      return self._declarations[item].default_value

    raise self.UnsetKeyError(
        'Configuration value not set and has no default', item)

  @threads.synchronized
  def __contains__(self, name):  # pylint: disable=invalid-name
    """True if we have a value for name."""
    return (name in self._declarations and
            (self._declarations[name].has_default or
             name in self._loaded_values or
             name in self._flag_values))

  @threads.synchronized
  def declare(self, name, description=None, **kwargs):
    """Declare a configuration key with the given name.

    Args:
      name: Configuration key to declare, must not have been already declared.
      description: If provided, use this as the description for this key.
      **kwargs: Other kwargs to pass to the Declaration, only default_value
          is currently supported.
    """
    if not self._is_valid_key(name):
      raise self.InvalidKeyError(
          'Invalid key name, must begin with a lowercase letter', name)
    if name in self._declarations:
      raise self.KeyAlreadyDeclaredError(
          'Configuration key already declared', name)
    self._declarations[name] = self.Declaration(
        name, description=description, **kwargs)

  @threads.synchronized
  def reset(self):
    """Reset the loaded state of the configuration to what it was at import.

    Note that this does *not* reset values set by commandline flags or loaded
    from --config-file (in fact, any values loaded from --config-file that have
    been overridden are reset to their value from --config-file).
    """
    # Populate loaded_values with values from --config-file, if it was given.
    self._loaded_values = {}
    if self._flags.config_file is not None:
      self.load_from_file(self._flags.config_file, _allow_undeclared=True)

  def load_from_filename(self, filename, *args, **kwargs):
    """Opens the filename and calls load_from_file."""
    with open(filename) as f:
      return self.load_from_file(f, *args, **kwargs)
  
  def load_from_file(self, yamlfile, _override=True, _allow_undeclared=False):
    """Loads the configuration from a file.

    Parsed contents must be a single dict mapping config key to value.

    Args:
      yamlfile: The opened file object to load configuration from.
      See load_from_dict() for other args' descriptions.

    Raises:
      ConfigurationInvalidError: If configuration file can't be read, or can't
          be parsed as either YAML (or JSON, which is a subset of YAML).
    """
    self._logger.info('Loading configuration from file: %s', yamlfile)

    try:
      parsed_yaml = self._modules['yaml'].safe_load(yamlfile.read())
    except self._modules['yaml'].YAMLError:
      self._logger.exception('Problem parsing YAML')
      raise self.ConfigurationInvalidError(
          'Failed to load from %s as YAML' % yamlfile)

    if not isinstance(parsed_yaml, dict):
      # Parsed YAML, but it's not a dict.
      raise self.ConfigurationInvalidError(
          'YAML parsed, but wrong type, should be dict', parsed_yaml)

    self._logger.debug('Configuration loaded from file: %s', parsed_yaml)
    self.load_from_dict(
        parsed_yaml, _override=_override, _allow_undeclared=_allow_undeclared)

  def load(self, _override=True, _allow_undeclared=False, **kwargs):
    """load configuration values from kwargs, see load_from_dict()."""
    self.load_from_dict(
        kwargs, _override=_override, _allow_undeclared=_allow_undeclared)

  @threads.synchronized
  def load_from_dict(self, dictionary, _override=True, _allow_undeclared=False):
    """Loads the config with values from a dictionary instead of a file.

    This is meant for testing and bin purposes and shouldn't be used in most
    applications.

    Args:
      dictionary: The dictionary containing config keys/values to update.
      _override: If True, new values will override previous values.
      _allow_undeclared: If True, silently load undeclared keys, otherwise
          warn and ignore the value.  Typically used for loading config
          files before declarations have been evaluated.
    """
    undeclared_keys = []
    for key, value in self._modules['six'].iteritems(dictionary):
      # Warn in this case.  We raise if you try to access a config key that
      # hasn't been declared, but we don't raise here so that you can use
      # configuration files that are supersets of required configuration for
      # any particular test station.
      if key not in self._declarations and not _allow_undeclared:
        undeclared_keys.append(key)
        continue
      if key in self._loaded_values:
        if _override:
          self._logger.info(
              'Overriding previously loaded value for %s (%s) with value: %s',
              key, self._loaded_values[key], value)
        else:
          self._logger.info(
              'Ignoring new value (%s), keeping previous value for %s: %s',
              value, key, self._loaded_values[key])
          continue

      # Force any keys and values that are bytes to unicode.
      key = key.decode() if isinstance(key, bytes) else key
      value = value.decode() if isinstance(value, bytes) else value

      self._loaded_values[key] = value
    if undeclared_keys:
      self._logger.warning('Ignoring undeclared configuration keys: %s',
                           undeclared_keys)

  @threads.synchronized
  def _asdict(self):
    """Create a dictionary snapshot of the current config values."""
    # Start with any default values we have, and override with loaded values,
    # and then override with flag values.
    retval = {key: self._declarations[key].default_value for
              key in self._declarations if self._declarations[key].has_default}
    retval.update(self._loaded_values)
    # Only update keys that are declared so we don't allow injecting
    # un-declared keys via commandline flags.
    for key, value in self._modules['six'].iteritems(self._flag_values):
      if key in self._declarations:
        retval[key] = value
    return retval

  @property
  def help_text(self):
    """Return a string with all config keys and their descriptions."""
    result = []
    for name in sorted(self._declarations.keys()):
      result.append(name)
      result.append('-' * len(name))
      decl = self._declarations[name]
      if decl.description:
        result.append(decl.description.strip())
      else:
        result.append('(no description found)')
      if decl.has_default:
        result.append('')
        quotes = '"' if type(decl.default_value) is str else ''
        result.append('  default_value={quotes}{val}{quotes}'.format(
            quotes=quotes, val=decl.default_value))
      result.append('')
      result.append('')
    return '\n'.join(result)

  def save_and_restore(self, _func=None, **config_values):
    """Decorator for saving conf state and restoring it after a function.

    This decorator is primarily for use in tests, where conf keys may be updated
    for individual test cases, but those values need to be reverted after the
    test case is done.

    Examples:

      conf.declare('my_conf_key')

      @conf.save_and_restore
      def MyTestFunc():
        conf.load(my_conf_key='baz')
        SomeFuncUnderTestThatUsesMyConfKey()

      conf.load(my_conf_key='foo')
      MyTestFunc()
      print conf.my_conf_key  # Prints 'foo', *NOT* 'baz'

      # Without the save_and_restore decorator, MyTestFunc() would have had the
      # side effect of altering the conf value of 'my_conf_key' to 'baz'.

      # Config keys can also be initialized for the context inline at decoration
      # time.  This is the same as setting them at the beginning of the
      # function, but is a little clearer syntax if you know ahead of time what
      # config keys and values you need to set.

      @conf.save_and_restore(my_conf_key='baz')
      def MyOtherTestFunc():
        print conf.my_conf_key  # Prints 'baz'

      MyOtherTestFunc()
      print conf.my_conf_key  # Prints 'foo' again, for the same reason.


    Args:
      _func: The function to wrap.  The returned wrapper will invoke the
          function and restore the config to the state it was in at invocation.
      **config_values: Config keys can be set inline at decoration time, see
          examples.  Note that config keys can't begin with underscore, so
          there can be no name collision with _func.

    Returns:
      Wrapper to replace _func, as per Python decorator semantics.
    """
    functools = self._modules['functools']  # pylint: disable=redefined-outer-name

    if not _func:
      return functools.partial(self.save_and_restore, **config_values)

    @functools.wraps(_func)
    def _saving_wrapper(*args, **kwargs):
      saved_config = dict(self._loaded_values)
      try:
        self.load_from_dict(config_values)
        return _func(*args, **kwargs)
      finally:
        self._loaded_values = saved_config # pylint: disable=attribute-defined-outside-init
    return _saving_wrapper

  def inject_positional_args(self, method):
    """Decorator for injecting positional arguments from the configuration.

    This decorator wraps the given method, so that any positional arguments are
    passed with corresponding values from the configuration.  The name of the
    positional argument must match the configuration key.

    Keyword arguments are *NEVER* modified, even if their names match
    configuration keys.  Avoid naming keyword args names that are also
    configuration keys to avoid confusion.

    Additional positional arguments may be used that do not appear in the
    configuration, but those arguments *MUST* be specified as keyword arguments
    upon invocation of the method.  This is to avoid ambiguity in which
    positional arguments are getting which values.

    Args:
      method: The method to wrap.

    Returns:
      A wrapper that, when invoked, will call the wrapped method, passing in
    configuration values for positional arguments.
    """
    inspect = self._modules['inspect']
    argspec = functions.getargspec(method)

    # Index in argspec.args of the first keyword argument.  This index is a
    # negative number if there are any kwargs, or 0 if there are no kwargs.
    keyword_arg_index = -1 * len(argspec.defaults or [])
    arg_names = argspec.args[:keyword_arg_index or None]
    kwarg_names = argspec.args[len(arg_names):]
    functools = self._modules['functools']  # pylint: disable=redefined-outer-name

    # Create the actual method wrapper, all we do is update kwargs.  Note we
    # don't pass any *args through because there can't be any - we've filled
    # them all in with values from the configuration.  Any positional args that
    # are missing from the configuration *must* be explicitly specified as
    # kwargs.
    @functools.wraps(method)
    def method_wrapper(**kwargs):
      """Wrapper that pulls values from openhtf.util.conf."""
      # Check for keyword args with names that are in the config so we can warn.
      for kwarg in kwarg_names:
        if kwarg in self:
          self._logger.warning('Keyword arg %s not set from configuration, but '
                               'is a configuration key', kwarg)

      # Set positional args from configuration values.
      final_kwargs = {name: self[name] for name in arg_names if name in self}

      for overridden in set(kwargs) & set(final_kwargs):
        self._logger.warning('Overriding configuration value for kwarg %s (%s) '
                             'with provided kwarg value: %s', overridden,
                             self[overridden], kwargs[overridden])

      final_kwargs.update(kwargs)
      if inspect.ismethod(method):
        name = '%s.%s' % (method.__self__.__class__.__name__, method.__name__)
      else:
        name = method.__name__
      self._logger.debug('Invoking %s with %s', name, final_kwargs)
      return method(**final_kwargs)

    # We have to check for a 'self' parameter explicitly because Python doesn't
    # pass it as a keyword arg, it passes it as the first positional arg.
    if argspec.args[0] == 'self':
      @functools.wraps(method)
      def self_wrapper(self, **kwargs):  # pylint: disable=invalid-name
        """Wrapper that pulls values from openhtf.util.conf."""
        kwargs['self'] = self
        return method_wrapper(**kwargs)
      return self_wrapper
    return method_wrapper

# Swap out the module for a singleton instance of Configuration so we can
# provide __getattr__ and __getitem__ functionality at the module level.
sys.modules[__name__] = Configuration(
    logging.getLogger(__name__), threading.RLock(), ARG_PARSER,
    functools=functools, inspect=inspect, yaml=yaml, six=six)



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020


"""PhaseExecutor module for handling the phases of a test.

Each phase is an instance of openhtf.PhaseDescriptor and therefore has
relevant options. Each option is taken into account when executing a phase,
such as checking options.run_if as soon as possible and timing out at the
appropriate time.

A phase must return an openhtf.PhaseResult, one of CONTINUE, REPEAT, or STOP.
A phase may also return None, or have no return statement, which is the same as
returning openhtf.PhaseResult.CONTINUE.  These results are then acted upon
accordingly and a new test run status is returned.

Phases are always run in order and not allowed to loop back, though a phase may
choose to repeat itself by returning REPEAT. Returning STOP will cause a test
to stop early, allowing a test to detect a bad state and not waste any further
time. A phase should not return TIMEOUT or ABORT, those are handled by the
framework.
"""

import collections
import logging
import sys
import threading
import time
import traceback

import openhtf
from openhtf.util import argv
from openhtf.util import threads
from openhtf.util import timeouts

DEFAULT_PHASE_TIMEOUT_S = 365 * 24 * 60 * 60

ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument(
    '--phase_default_timeout_s', default=DEFAULT_PHASE_TIMEOUT_S,
    action=argv.StoreInModule, target='%s.DEFAULT_PHASE_TIMEOUT_S' % __name__,
    help='Test phase timeout in seconds')

_LOG = logging.getLogger(__name__)


class ExceptionInfo(collections.namedtuple(
    'ExceptionInfo', ['exc_type', 'exc_val', 'exc_tb'])):
  """Wrap the description of a raised exception and its traceback."""

  def _asdict(self):
    return {
        'exc_type': str(self.exc_type),
        'exc_val': self.exc_val,
        'exc_tb': ''.join(traceback.format_exception(*self)),
    }

  def __str__(self):
    return self.exc_type.__name__


class InvalidPhaseResultError(Exception):
  """Raised when PhaseExecutionOutcome is created with invalid phase result."""


class PhaseExecutionOutcome(collections.namedtuple(
    'PhaseExecutionOutcome', 'phase_result')):
  """Provide some utility and sanity around phase return values.

  This should not be confused with openhtf.PhaseResult.  PhaseResult is an
  enumeration to provide user-facing valid phase return values.  This tuple
  is used internally to track other possible outcomes (timeout, exception),
  and to perform some sanity checking (weird return values from phases).

  If phase_result is None, that indicates the phase timed out (this makes
  sense if you think about it, it timed out, so there was no result).  If
  phase_result is an instance of Exception, then that is the Exception that
  was raised by the phase.  The raised_exception attribute can be used as
  a convenience to test for that condition, and the is_timeout attribute can
  similarly be used to check for the timeout case.

  The only accepted values for phase_result are None (timeout), an instance
  of Exception (phase raised), or an instance of openhtf.PhaseResult.  Any
  other value will raise an InvalidPhaseResultError.
  """

  def __new__(cls, phase_result):
    if (phase_result is not None and
        not isinstance(phase_result, (openhtf.PhaseResult, ExceptionInfo)) and
        not isinstance(phase_result, threads.ThreadTerminationError)):
      raise InvalidPhaseResultError('Invalid phase result', phase_result)
    self = super(PhaseExecutionOutcome, cls).__new__(cls, phase_result)
    return self

  @property
  def is_fail_and_continue(self):
    return self.phase_result is openhtf.PhaseResult.FAIL_AND_CONTINUE

  @property
  def is_repeat(self):
    return self.phase_result is openhtf.PhaseResult.REPEAT

  @property
  def is_skip(self):
    return self.phase_result is openhtf.PhaseResult.SKIP

  @property
  def is_terminal(self):
    """True if this result will stop the test."""
    return (self.raised_exception or self.is_timeout or
            self.phase_result == openhtf.PhaseResult.STOP)

  @property
  def is_timeout(self):
    """True if this PhaseExecutionOutcome indicates a phase timeout."""
    return self.phase_result is None

  @property
  def raised_exception(self):
    """True if the phase in question raised an exception."""
    return isinstance(self.phase_result, (
        ExceptionInfo, threads.ThreadTerminationError))


class PhaseExecutorThread(threads.KillableThread):
  """Handles the execution and result of a single test phase.

  The phase outcome will be stored in the _phase_execution_outcome attribute
  once it is known (_phase_execution_outcome is None until then), and it will be
  a PhaseExecutionOutcome instance.
  """
  daemon = True

  def __init__(self, phase_desc, test_state):
    super(PhaseExecutorThread, self).__init__(
        name='<PhaseExecutorThread: (phase_desc.name)>')
    self._phase_desc = phase_desc
    self._test_state = test_state
    self._phase_execution_outcome = None

  def _thread_proc(self):
    """Execute the encompassed phase and save the result."""
    # Call the phase, save the return value, or default it to CONTINUE.
    phase_return = self._phase_desc(self._test_state)
    if phase_return is None:
      phase_return = openhtf.PhaseResult.CONTINUE

    # If phase_return is invalid, this will raise, and _phase_execution_outcome
    # will get set to the InvalidPhaseResultError in _thread_exception instead.
    self._phase_execution_outcome = PhaseExecutionOutcome(phase_return)

  def _log_exception(self, *args):
    """Log exception, while allowing unit testing to override."""
    self._test_state.state_logger.critical(*args)

  def _thread_exception(self, *args):
    self._phase_execution_outcome = PhaseExecutionOutcome(ExceptionInfo(*args))
    self._log_exception('Phase %s raised an exception', self._phase_desc.name)
    return True  # Never propagate exceptions upward.

  def join_or_die(self):
    """Wait for thread to finish, returning a PhaseExecutionOutcome instance."""
    if self._phase_desc.options.timeout_s is not None:
      threads.join_timeout_increments(self, self._phase_desc.options.timeout_s)
    else:
      threads.join_timeout_increments(self, DEFAULT_PHASE_TIMEOUT_S)

    # We got a return value or an exception and handled it.
    if isinstance(self._phase_execution_outcome, PhaseExecutionOutcome):
      return self._phase_execution_outcome

    # Check for timeout, indicated by None for
    # PhaseExecutionOutcome.phase_result.
    if self.is_alive():
      self.kill()
      return PhaseExecutionOutcome(None)

    # Phase was killed.
    return PhaseExecutionOutcome(threads.ThreadTerminationError())

  @property
  def name(self):
    return str(self)

  def __str__(self):
    return '<%s: (%s)>' % (type(self).__name__, self._phase_desc.name)


class PhaseExecutor(object):
  """Encompasses the execution of the phases of a test."""

  def __init__(self, test_state):
    self.test_state = test_state
    # This lock exists to prevent stop() calls from being ignored if called when
    # _execute_phase_once is setting up the next phase thread.
    self._current_phase_thread_lock = threading.Lock()
    self._current_phase_thread = None
    self._stopping = threading.Event()

  def execute_phase(self, phase):
    """Executes a phase or skips it, yielding PhaseExecutionOutcome instances.

    Args:
      phase: Phase to execute.

    Returns:
      The final PhaseExecutionOutcome that wraps the phase return value
      (or exception) of the final phase run. All intermediary results, if any,
      are REPEAT and handled internally. Returning REPEAT here means the phase
      hit its limit for repetitions.
    """
    repeat_count = 1
    repeat_limit = phase.options.repeat_limit or sys.maxsize
    while not self._stopping.is_set():
      is_last_repeat = repeat_count >= repeat_limit
      phase_execution_outcome = self._execute_phase_once(phase, is_last_repeat)

      if phase_execution_outcome.is_repeat and not is_last_repeat:
        repeat_count += 1
        continue

      return phase_execution_outcome
    # We've been cancelled, so just 'timeout' the phase.
    return PhaseExecutionOutcome(None)

  def _execute_phase_once(self, phase_desc, is_last_repeat):
    """Executes the given phase, returning a PhaseExecutionOutcome."""
    # Check this before we create a PhaseState and PhaseRecord.
    if not phase_desc.options.call_run_if(self.test_state.user_defined_state):
      _LOG.debug('Phase %s skipped due to run_if returning falsey.',
                 phase_desc.name)
      return PhaseExecutionOutcome(openhtf.PhaseResult.SKIP)

    override_result = None
    with self.test_state.running_phase_context(phase_desc) as phase_state:
      _LOG.debug('Executing phase %s', phase_desc.name)
      with self._current_phase_thread_lock:
        # Checking _stopping must be in the lock context, otherwise there is a
        # race condition: this thread checks _stopping and then switches to
        # another thread where stop() sets _stopping and checks
        # _current_phase_thread (which would not be set yet).  In that case, the
        # new phase thread will be still be started.
        if self._stopping.is_set():
          # PhaseRecord will be written at this point, so ensure that it has a
          # Killed result.
          result = PhaseExecutionOutcome(threads.ThreadTerminationError())
          phase_state.result = result
          return result
        phase_thread = PhaseExecutorThread(phase_desc, self.test_state)
        phase_thread.start()
        self._current_phase_thread = phase_thread

      phase_state.result = phase_thread.join_or_die()
      if phase_state.result.is_repeat and is_last_repeat:
        _LOG.error('Phase returned REPEAT, exceeding repeat_limit.')
        phase_state.hit_repeat_limit = True
        override_result = PhaseExecutionOutcome(openhtf.PhaseResult.STOP)
      self._current_phase_thread = None

    # Refresh the result in case a validation for a partially set measurement
    # raised an exception.
    result = override_result or phase_state.result
    _LOG.debug('Phase %s finished with result %s', phase_desc.name,
               result.phase_result)
    return result

  def reset_stop(self):
    self._stopping.clear()

  def stop(self, timeout_s=None):
    """Stops execution of the current phase, if any.

    It will raise a ThreadTerminationError, which will cause the test to stop
    executing and terminate with an ERROR state.

    Args:
      timeout_s: int or None, timeout in seconds to wait for the phase to stop.
    """
    self._stopping.set()
    with self._current_phase_thread_lock:
      phase_thread = self._current_phase_thread
      if not phase_thread:
        return

    if phase_thread.is_alive():
      phase_thread.kill()

      _LOG.debug('Waiting for cancelled phase to exit: %s', phase_thread)
      timeout = timeouts.PolledTimeout.from_seconds(timeout_s)
      while phase_thread.is_alive() and not timeout.has_expired():
        time.sleep(0.1)
      _LOG.debug('Cancelled phase %s exit',
                 "didn't" if phase_thread.is_alive() else 'did')
    # Clear the currently running phase, whether it finished or timed out.
    self.test_state.stop_running_phase()



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Some handy interfaces to the ADB :shell service.

The :shell service is pretty straightforward, you send 'shell:command' and
the device runs /bin/sh -c 'command'.  The ADB daemon on the device sets up a
PTY, similar to what ssh would do, to provide interactive terminal features.
This makes things difficult if you're trying to pipe binary data through a
remote command; however, we provide some facilities for 'raw' commands, where
we first execute an ioctl to turn off things like character translation and
local echo, providing a more sane programmatic interface.

The ShellService also provides some handy methods for running commands
asynchronously, either by returning a handle, or using a with: context.

Note that this service differs from the FilesyncService in that streams opened
to the :shell service are closed immediately after the command completes;
streams opened to the :sync service remain open for multiple sync requests,
until they are closed explicitly.  This means there's no point in keeping a
stream around ShellService, we need to keep an AdbConnection around instead.

Some examples of how to use this service:
  adb_cnxn = adb_protocol.AdbConnection.connect(my_transport)
  shell = shell_service.ShellService(adb_cnxn)

  # Run a simple command.
  output = shell.command('echo foo')
  # output == 'foo\r\n'

  # Run a command that outputs binary data, like recording a minute of audio.
  output = shell.RawCommand('arecord -Dhw:CARD=0,DEV=0 -c 2 -d 60')

  # Run a command in the background, do some other stuff, then read the
  # command's output, waiting on it to complete.
  cmd = shell.async_command('echo foo; sleep 10')
  bar = shell.command('echo bar')
  foo = cmd.wait()
  baz = shell.command('echo baz')

  # A version using a with context to do the same thing:
  with shell.async_command('echo foo; sleep 10') as c:
    bar = shell.command('echo bar')
    foo = c.wait()
  baz = shell.command('echo baz')

  # Run a command in the background while we do some other stuff, save the
  # output to a StringIO buffer so we can access it later.  Use a context to
  # automatically wait for the asynchronous command to finish.
  output = cStringIO.StringIO()
  with shell.AsyncRawCommand(
      'arecord -Dhw:CARD=0,DEV=0 -c 2 -d 60', stdout=output):
    # Do some stuff, play some sounds on some fixture speakers, for example.
    pass
  # Execution won't get here until the arecord command completes, and
  # output.getvalue() now contains the output of the arecord command.
"""

import threading
import time

from openhtf.plugs.usb import adb_protocol
from openhtf.plugs.usb import usb_exceptions

from openhtf.util import timeouts
import six


class AsyncCommandHandle(object):
  """This class is used for interacting with an asynchronous command.

  This handle is used to close a command or to wait on it to complete. Data is
  read from stdin and written to the command's stdin, and output from the
  command is written to stdout. If stdin is None, no input is written to the
  command.  If stdout is None, the output from the command is buffered
  internally, and will be returned from a call to wait() - see the wait() method
  for details.

  You can tell if a stream was closed locally by checking the
  'force_closed_or_timeout' attribute.  If a command completes instead of being
  closed by a call to Close (or a timeout), then 'force_closed_or_timeout' will
  be False, otherwise it will be True.
  """

  def __init__(self, stream, stdin, stdout, timeout, is_raw):  #pylint: disable=too-many-arguments
    """Create a handle to use for interfacing with an async_command.

    Args:
      stream: Stream to use for communicating with the running command.
      stdin: File-like object to use for reading stdin for the command, can be
        None, in which case no input is sent to the command.
      stdout: File-like object to use for writing output of the command to, can
        be None, in which case output can be obtained by calling wait().
      timeout: timeouts.PolledTimeout to use for the command.
      is_raw: If True, we'll do reads from stdin, otherwise we do readlines
        instead to play nicer with potential interactive uses (read doesn't
        return until EOF, but interactively you want to send each line and
        then see the response).  stdout is treated the same in either case,
        read is used - AdbStreams don't support readline.
    """
    self.stream = stream
    self.stdin = stdin
    self.stdout = stdout or six.StringIO()
    self.force_closed_or_timeout = False

    self.reader_thread = threading.Thread(target=self._reader_thread_proc,
                                          args=(timeout,))
    self.reader_thread.daemon = True
    self.reader_thread.start()

    if stdin:
      self.writer_thread = threading.Thread(target=self._writer_thread_proc,
                                            args=(is_raw,))
      self.writer_thread.daemon = True
      self.writer_thread.start()

    # Close ourselves after timeout expires, ignored if timeout won't expire.
    timeouts.execute_after_delay(timeout, self.close)

  def _writer_thread_proc(self, is_raw):
    """Write as long as the stream is not closed."""
    # If we're not in raw mode, do line-buffered reads to play nicer with
    # potential interactive uses, max of MAX_ADB_DATA, since anything we write
    # to the stream will get packetized to that size anyway.
    #
    # Loop until our stream gets closed, which will cause one of these
    # operations to raise.  Since we're in a separate thread, it'll just get
    # ignored, which is what we want.
    reader = self.stdin.read if is_raw else self.stdin.readline
    while not self.stream.is_closed():
      self.stream.write(reader(adb_protocol.MAX_ADB_DATA))

  def _reader_thread_proc(self, timeout):
    """Read until the stream is closed."""
    for data in self.stream.read_until_close(timeout_ms=timeout):
      if self.stdout is not None:
        self.stdout.write(data)

  def __enter__(self):  # pylint: disable=invalid-name
    return self

  def __exit__(self, exc_type, exc_value, exc_tb):  # pylint: disable=invalid-name
    if exc_type:
      return False
    self.wait()
    return True

  def close(self):
    """Close this handle immediately - you may lose output."""
    self.force_closed_or_timeout = True
    self.stream.close()

  def is_done(self):
    """Return True if this command has completed."""
    return self.stream.is_closed()

  def wait(self, timeout_ms=None):
    """Block until this command has completed.

    Args:
      timeout_ms: Timeout, in milliseconds, to wait.

    Returns:
      Output of the command if it complete and self.stdout is a StringIO
    object or was passed in as None.  Returns True if the command completed but
    stdout was provided (and was not a StringIO object).  Returns None if the
    timeout expired before the command completed.  Be careful to check the
    return value explicitly for None, as the output may be ''.
    """
    closed = timeouts.loop_until_timeout_or_true(
        timeouts.PolledTimeout.from_millis(timeout_ms),
        self.stream.is_closed, .1)
    if closed:
      if hasattr(self.stdout, 'getvalue'):
        return self.stdout.getvalue()
      return True
    return None


class ShellService(object):
  """Class providing a high-level interface to ADB's :shell service.

  This class provides synchronous and asynchronous commands, and a variety of
  ways for getting input into and out of them.
  """

  def __init__(self, adb_connection):
    self.adb_connection = adb_connection

  @staticmethod
  def _to_raw_command(command):
    """Convert the command to a raw signal."""
    # Android doesn't have stty, so we manually do the ioctl (yuck).  This ioctl
    # is a TCSETA (0x5403) with the following flags set:
    #   Control bits:
    #     B38400 (set baud rate)
    #     CS8 (8-bit bytes)
    #     CREAD (Enable input from terminal)
    #   Input, Output, Local bits all cleared
    #
    # We also update VMIN from 0x0 to 0xff so read() waits for at least one byte
    # to be ready before returning (we leave the default VTIME at 0x4).  Note
    # that we leave the other control characters at their defaults, but they
    # should be ignored since we disable them with flags and put the terminal
    # into non-canonical input mode (not newline delimited).
    return ('ioctl -l 23 -a 1 /proc/self/fd/0 0x5403 '  # TCSETA (0x5403)
            '0 0 0 0 0 0 0 0 0xbf 0 0 0 0 0 0 0 '  # Flags
            '0 0x3 0x1c 0x7f 0x15 0x4 0xff '  # Control characters
            '&>/dev/null;%s' % command)

  def command(self, command, raw=False, timeout_ms=None):
    """Run the given command and return the output."""
    return ''.join(self.streaming_command(command, raw, timeout_ms))

  def streaming_command(self, command, raw=False, timeout_ms=None):
    """Run the given command and yield the output as we receive it."""
    if raw:
      command = self._to_raw_command(command)
    return self.adb_connection.streaming_command('shell', command, timeout_ms)

  # pylint: disable=too-many-arguments
  def async_command(self, command, stdin=None, stdout=None, raw=False,
                    timeout_ms=None):
    """Run the given command on the device asynchronously.

    Input will be read from stdin, output written to stdout.  ADB doesn't
    distinguish between stdout and stdin on the device, so they get interleaved
    into stdout here.  stdin and stdout should be file-like objects, so you
    could use sys.stdin and sys.stdout to emulate the 'adb shell' commandline.

    Args:
      command: The command to run, will be run with /bin/sh -c 'command' on
        the device.
      stdin: File-like object to read from to pipe to the command's stdin.  Can
        be None, in which case nothing will be written to the command's stdin.
      stdout: File-like object to write the command's output to.  Can be None,
        in which case the command's output will be buffered internally, and can
        be access via the return value of wait().
      raw: If True, run the command as per RawCommand (see above).
      timeout_ms: Timeout for the command, in milliseconds.

    Returns:
      An AsyncCommandHandle instance that can be used to send/receive data to
    and from the command or wait on the command to finish.

    Raises:
      AdbStreamUnavailableError: If the remote devices doesn't support the
        shell: service.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    if raw:
      command = self._to_raw_command(command)
    stream = self.adb_connection.open_stream('shell:%s' % command, timeout)
    if not stream:
      raise usb_exceptions.AdbStreamUnavailableError(
          '%s does not support service: shell', self)
    if raw and stdin is not None:
      # Short delay to make sure the ioctl to set raw mode happens before we do
      # any writes to the stream, if we don't do this bad things happen...
      time.sleep(.1)
    return AsyncCommandHandle(stream, stdin, stdout, timeout, raw)
  # pylint: enable=too-many-arguments

  @classmethod
  def using_connection(cls, adb_connection):
    """Factory method to match the interface of FilesyncService."""
    return cls(adb_connection)



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The plugs module provides boilerplate for accessing hardware.

Most tests require interaction with external hardware.  This module provides
framework support for such interfaces, allowing for automatic setup and
teardown of the objects.

Test phases can be decorated as using Plug objects, which then get passed
into the test via parameters.  Plugs are all instantiated at the
beginning of a test, and all plugs' tearDown() methods are called at the
end of a test.  It's up to the Plug implementation to do any sort of
is-ready check.

A plug may be made "frontend-aware", allowing it, in conjunction with the
Station API, to update any frontends each time the plug's state changes. See
FrontendAwareBasePlug for more info.

Example implementation of a plug:

  from openhtf import plugs

  class ExamplePlug(plugs.BasePlug):
    '''A Plug that does nothing.'''

    def __init__(self):
      print 'Instantiating %s!' % type(self).__name__

    def DoSomething(self):
      print '%s doing something!' % type(self).__name__

    def tearDown(self):
      # This method is optional.  If implemented, it will be called at the end
      # of the test.
      print 'Tearing down %s!' % type(self).__name__

Example usage of the above plug:

  from openhtf import plugs
  from my_custom_plugs_package import example

  @plugs.plug(example=example.ExamplePlug)
  def TestPhase(test, example):
    print 'Test phase started!'
    example.DoSomething()
    print 'Test phase done!'

Putting all this together, when the test is run (with just that phase), you
would see the output (with other framework logs before and after):

  Instantiating ExamplePlug!
  Test phase started!
  ExamplePlug doing something!
  Test phase done!
  Tearing down ExamplePlug!

Plugs will often need to use configuration values.  The recommended way
of doing this is with the conf.inject_positional_args decorator:

  from openhtf import plugs
  from openhtf.util import conf

  conf.declare('my_config_key', default_value='my_config_value')

  class ExamplePlug(plugs.BasePlug):
    '''A plug that requires some configuration.'''

    @conf.inject_positional_args
    def __init__(self, my_config_key)
      self._my_config = my_config_key

Note that Plug constructors shouldn't take any other arguments; the
framework won't pass any, so you'll get a TypeError.  Any values that are only
known at run time must be either passed into other methods or set via explicit
setter methods.  See openhtf/conf.py for details, but with the above
example, you would also need a configuration .yaml file with something like:

  my_config_key: my_config_value

This will result in the ExamplePlug being constructed with
self._my_config having a value of 'my_config_value'.
"""

import collections
import logging

import mutablerecords

from openhtf import util
import openhtf.core.phase_descriptor
from openhtf.util import classproperty
from openhtf.util import conf
from openhtf.util import data
from openhtf.util import logs
from openhtf.util import threads
import six


_LOG = logging.getLogger(__name__)


conf.declare('plug_teardown_timeout_s', default_value=0, description=
             'Timeout (in seconds) for each plug tearDown function if > 0; '
             'otherwise, will wait an unlimited time.')


PlugDescriptor = collections.namedtuple('PlugDescriptor', ['mro'])  # pylint: disable=invalid-name

# Placeholder for a specific plug to be provided before test execution.
#
# Use the with_plugs() method to provide the plug before test execution. The
# with_plugs() method checks to make sure the substitute plug is a subclass of
# the PlugPlaceholder's base_class.
PlugPlaceholder = collections.namedtuple('PlugPlaceholder', ['base_class'])  # pylint: disable=invalid-name


class PhasePlug(mutablerecords.Record(
    'PhasePlug', ['name', 'cls'], {'update_kwargs': True})):
  """Information about the use of a plug in a phase."""


class PlugOverrideError(Exception):
  """Raised when a plug would be overridden by a kwarg."""


class DuplicatePlugError(Exception):
  """Raised when the same plug is required multiple times on a phase."""


class InvalidPlugError(Exception):
  """Raised when a plug declaration or requested name is invalid."""


class BasePlug(object):
  """All plug types must subclass this type.

  Attributes:
    logger: This attribute will be set by the PlugManager (and as such it
        doesn't appear here), and is the same logger as passed into test
        phases via TestApi.
  """
  # Override this to True in subclasses to support remote Plug access.
  enable_remote = False
  # Allow explicitly disabling remote access to specific attributes.
  disable_remote_attrs = set()
  # Override this to True in subclasses to support using with_plugs with this
  # plug without needing to use placeholder.  This will only affect the classes
  # that explicitly define this; subclasses do not share the declaration.
  auto_placeholder = False
  # Default logger to be used only in __init__ of subclasses.
  # This is overwritten both on the class and the instance so don't store
  # a copy of it anywhere.
  logger = _LOG

  @classproperty
  def placeholder(cls):
    """Returns a PlugPlaceholder for the calling class."""
    return PlugPlaceholder(cls)

  def _asdict(self):
    """Return a dictionary representation of this plug's state.

    This is called repeatedly during phase execution on any plugs that are in
    use by that phase.  The result is reported via the Station API by the
    PlugManager (if the Station API is enabled, which is the default).

    Note that this method is called in a tight loop, it is recommended that you
    decorate it with functions.call_at_most_every() to limit the frequency at
    which updates happen (pass a number of seconds to it to limit samples to
    once per that number of seconds).

    You can also implement an `as_base_types` function that can return a dict
    where the values must be base types at all levels.  This can help prevent
    recursive copying, which is time intensive.
    """
    return {}

  def tearDown(self):
    """This method is called automatically at the end of each Test execution."""
    pass

  @classmethod
  def uses_base_tear_down(cls):
    """Checks whether the tearDown method is the BasePlug implementation."""
    this_tear_down = getattr(cls, 'tearDown')
    base_tear_down = getattr(BasePlug, 'tearDown')
    return this_tear_down.__code__ is base_tear_down.__code__


class FrontendAwareBasePlug(BasePlug, util.SubscribableStateMixin):
  """A plug that notifies of any state updates.

  Plugs inheriting from this class may be used in conjunction with the Station
  API to update any frontends each time the plug's state changes. The plug
  should call notify_update() when and only when the state returned by _asdict()
  changes.

  Since the Station API runs in a separate thread, the _asdict() method of
  frontend-aware plugs should be written with thread safety in mind.
  """
  enable_remote = True


def plug(update_kwargs=True, **plugs_map):
  """Creates a decorator that passes in plugs when invoked.

  This function returns a decorator for a function that will replace positional
  arguments to that function with the plugs specified.  See the module
  docstring for details and examples.

  Note this decorator does not work with class or bound methods, but does work
  with @staticmethod.

  Args:
    update_kwargs: If true, makes the decorated phase take this plug as a kwarg.
    **plugs_map: Dict mapping name to Plug type.

  Returns:
    A PhaseDescriptor that will pass plug instances in as kwargs when invoked.

  Raises:
    InvalidPlugError: If a type is provided that is not a subclass of BasePlug.
  """
  for a_plug in plugs_map.values():
    if not (isinstance(a_plug, PlugPlaceholder)
            or issubclass(a_plug, BasePlug)):
      raise InvalidPlugError(
          'Plug %s is not a subclass of plugs.BasePlug nor a placeholder '
          'for one' % a_plug)

  def result(func):
    """Wrap the given function and return the wrapper.

    Args:
      func: The function to wrap.

    Returns:
      A PhaseDescriptor that, when called will invoke the wrapped function,
        passing plugs as keyword args.

    Raises:
      DuplicatePlugError:  If a plug name is declared twice for the
          same function.
    """
    phase = openhtf.core.phase_descriptor.PhaseDescriptor.wrap_or_copy(func)
    duplicates = (frozenset(p.name for p in phase.plugs) &
                  frozenset(plugs_map))
    if duplicates:
      raise DuplicatePlugError(
          'Plugs %s required multiple times on phase %s' % (duplicates, func))

    phase.plugs.extend([
        PhasePlug(name, a_plug, update_kwargs=update_kwargs)
        for name, a_plug in six.iteritems(plugs_map)])
    return phase
  return result


class _PlugTearDownThread(threads.KillableThread):
  """Killable thread that runs a plug's tearDown function."""

  def __init__(self, a_plug, *args, **kwargs):
    super(_PlugTearDownThread, self).__init__(*args, **kwargs)
    self._plug = a_plug

  def _thread_proc(self):
    try:
      self._plug.tearDown()
    except Exception:  # pylint: disable=broad-except
      # Including the stack trace from ThreadTerminationErrors received when
      # killed.
      _LOG.warning('Exception calling tearDown on %s:',
                   self._plug, exc_info=True)


class PlugManager(object):
  """Class to manage the lifetimes of plugs.

  This class handles instantiation of plugs at test start and calling
  tearDown() on all plugs when the test completes.  It is used by
  the executor, and should not be instantiated outside the framework itself.

  Note this class is not thread-safe.  It should only ever be used by the
  main framework thread anyway.

  Attributes:
    _plug_types: Initial set of plug types, additional plug types may be
        passed into calls to initialize_plugs().
    _plugs_by_type: Dict mapping plug type to plug instance.
    _plugs_by_name: Dict mapping plug name to plug instance.
    _plug_descriptors: Dict mapping plug type to plug descriptor.
    logger: logging.Logger instance that can save logs to the running test
        record.
  """

  def __init__(self, plug_types=None, record_logger=None):
    self._plug_types = plug_types or set()
    for plug in self._plug_types:
      if isinstance(plug, PlugPlaceholder):
        raise InvalidPlugError('Plug %s is a placeholder, replace it using '
                               'with_plugs().' % plug)
    self._plugs_by_type = {}
    self._plugs_by_name = {}
    self._plug_descriptors = {}
    if not record_logger:
      record_logger = _LOG
    self.logger = record_logger.getChild('plug')

  def as_base_types(self):
    return {
        'plug_descriptors': {
            name: dict(descriptor._asdict())  # Convert OrderedDict to dict.
            for name, descriptor in six.iteritems(self._plug_descriptors)
        },
        'plug_states': {
            name: data.convert_to_base_types(plug)
            for name, plug in six.iteritems(self._plugs_by_name)
        },
    }

  def _make_plug_descriptor(self, plug_type):
    """Returns the plug descriptor, containing info about this plug type."""
    return PlugDescriptor(self.get_plug_mro(plug_type))

  def get_plug_mro(self, plug_type):
    """Returns a list of names identifying the plug classes in the plug's MRO.

    For example:
        ['openhtf.plugs.user_input.UserInput']
    Or:
        ['openhtf.plugs.user_input.UserInput',
         'my_module.advanced_user_input.AdvancedUserInput']
    """
    ignored_classes = (BasePlug, FrontendAwareBasePlug)
    return [
        self.get_plug_name(base_class) for base_class in plug_type.mro()
        if (issubclass(base_class, BasePlug) and
            base_class not in ignored_classes)
    ]

  def get_plug_name(self, plug_type):
    """Returns the plug's name, which is the class name and module.

    For example:
        'openhtf.plugs.user_input.UserInput'
    """
    return '%s.%s' % (plug_type.__module__, plug_type.__name__)

  def initialize_plugs(self, plug_types=None):
    """Instantiate required plugs.

    Instantiates plug types and saves the instances in self._plugs_by_type for
    use in provide_plugs().

    Args:
      plug_types: Plug types may be specified here rather than passed
                  into the constructor (this is used primarily for unit testing
                  phases).
    """
    types = plug_types if plug_types is not None else self._plug_types
    for plug_type in types:
      # Create a logger for this plug. All plug loggers go under the 'plug'
      # sub-logger in the logger hierarchy.
      plug_logger = self.logger.getChild(plug_type.__name__)
      if plug_type in self._plugs_by_type:
        continue
      try:
        if not issubclass(plug_type, BasePlug):
          raise InvalidPlugError(
              'Plug type "%s" is not an instance of BasePlug' % plug_type)
        if plug_type.logger != _LOG:
          # They put a logger attribute on the class itself, overriding ours.
          raise InvalidPlugError(
              'Do not override "logger" in your plugs.', plug_type)

        # Override the logger so that __init__'s logging goes into the record.
        plug_type.logger = plug_logger
        try:
          plug_instance = plug_type()
        finally:
          # Now set it back since we'll give the instance a logger in a moment.
          plug_type.logger = _LOG
        # Set the logger attribute directly (rather than in BasePlug) so we
        # don't depend on subclasses' implementation of __init__ to have it
        # set.
        if plug_instance.logger != _LOG:
          raise InvalidPlugError(
              'Do not set "self.logger" in __init__ in your plugs', plug_type)
        else:
          # Now the instance has its own copy of the test logger.
          plug_instance.logger = plug_logger
      except Exception:  # pylint: disable=broad-except
        plug_logger.exception('Exception instantiating plug type %s', plug_type)
        self.tear_down_plugs()
        raise
      self.update_plug(plug_type, plug_instance)

  def get_plug_by_class_path(self, plug_name):
    """Get a plug instance by name (class path).

    This provides a way for extensions to OpenHTF to access plug instances for
    a running test via that test's plug manager.

    Args:
      plug_name: Plug name, e.g. 'openhtf.plugs.user_input.UserInput'.

    Returns:
      The plug manager's instance of the specified plug.
    """
    return self._plugs_by_name.get(plug_name)

  def update_plug(self, plug_type, plug_value):
    """Update internal data stores with the given plug value for plug type.

    Safely tears down the old instance if one was already created, but that's
    generally not the case outside unittests.  Also, we explicitly pass the
    plug_type rather than detecting it from plug_value to allow unittests to
    override plugs with Mock instances.

    Note this should only be used inside unittests, as this mechanism is not
    compatible with RemotePlug support.
    """
    self._plug_types.add(plug_type)
    if plug_type in self._plugs_by_type:
      self._plugs_by_type[plug_type].tearDown()
    plug_name = self.get_plug_name(plug_type)
    self._plugs_by_type[plug_type] = plug_value
    self._plugs_by_name[plug_name] = plug_value
    self._plug_descriptors[plug_name] = self._make_plug_descriptor(plug_type)

  def provide_plugs(self, plug_name_map):
    """Provide the requested plugs [(name, type),] as {name: plug instance}."""
    return {name: self._plugs_by_type[cls] for name, cls in plug_name_map}

  def tear_down_plugs(self):
    """Call tearDown() on all instantiated plugs.

    Note that initialize_plugs must have been called before calling
    this method, and initialize_plugs must be called again after calling
    this method if you want to access the plugs attribute again.

    Any exceptions in tearDown() methods are logged, but do not get raised
    by this method.
    """
    _LOG.debug('Tearing down all plugs.')
    for plug_type, plug_instance in six.iteritems(self._plugs_by_type):
      if plug_instance.uses_base_tear_down():
        name = '<PlugTearDownThread: BasePlug No-Op for %s>' % plug_type
      else:
        name = '<PlugTearDownThread: %s>' % plug_type
      thread = _PlugTearDownThread(plug_instance, name=name)
      thread.start()
      timeout_s = (conf.plug_teardown_timeout_s
                   if conf.plug_teardown_timeout_s
                   else None)
      thread.join(timeout_s)
      if thread.is_alive():
        thread.kill()
        _LOG.warning('Killed tearDown for plug %s after timeout.',
                     plug_instance)
    self._plugs_by_type.clear()
    self._plugs_by_name.clear()

  def wait_for_plug_update(self, plug_name, remote_state, timeout_s):
    """Wait for a change in the state of a frontend-aware plug.

    Args:
      plug_name: Plug name, e.g. 'openhtf.plugs.user_input.UserInput'.
      remote_state: The last observed state.
      timeout_s: Number of seconds to wait for an update.

    Returns:
      An updated state, or None if the timeout runs out.

    Raises:
      InvalidPlugError: The plug can't be waited on either because it's not in
          use or it's not a frontend-aware plug.
    """
    plug = self._plugs_by_name.get(plug_name)

    if plug is None:
      raise InvalidPlugError('Cannot wait on unknown plug "%s".' % plug_name)

    if not isinstance(plug, FrontendAwareBasePlug):
      raise InvalidPlugError('Cannot wait on a plug %s that is not an subclass '
                             'of FrontendAwareBasePlug.' % plug_name)

    state, update_event = plug.asdict_with_event()
    if state != remote_state:
      return state

    if update_event.wait(timeout_s):
      return plug._asdict()

  def get_frontend_aware_plug_names(self):
    """Returns the names of frontend-aware plugs."""
    return [name for name, plug in six.iteritems(self._plugs_by_name)
            if isinstance(plug, FrontendAwareBasePlug)]



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for functions."""

import collections
import functools
import inspect
import time


def getargspec(func):
  fullargspec = inspect.getfullargspec(func)
  return inspect.ArgSpec(
    args=fullargspec.args, 
    varargs=fullargspec.varargs,
    keywords=fullargspec.varkw,
    defaults=fullargspec.defaults)
  

def call_once(func):
  """Decorate a function to only allow it to be called once.

  Note that it doesn't make sense to only call a function once if it takes
  arguments (use @functools.lru_cache for that sort of thing), so this only
  works on callables that take no args.
  """
  argspec = getargspec(func)
  if argspec.args or argspec.varargs or argspec.keywords:
    raise ValueError('Can only decorate functions with no args', func, argspec)

  @functools.wraps(func)
  def _wrapper():
    # If we haven't been called yet, actually invoke func and save the result.
    if not _wrapper.HasRun():
      _wrapper.MarkAsRun()
      _wrapper.return_value = func()
    return _wrapper.return_value

  _wrapper.has_run = False
  _wrapper.HasRun = lambda: _wrapper.has_run
  _wrapper.MarkAsRun = lambda: setattr(_wrapper, 'has_run', True)
  return _wrapper

def call_at_most_every(seconds, count=1):
  """Call the decorated function at most count times every seconds seconds.

  The decorated function will sleep to ensure that at most count invocations
  occur within any 'seconds' second window.
  """
  def decorator(func):
    try:
      call_history = getattr(func, '_call_history')
    except AttributeError:
      call_history = collections.deque(maxlen=count)
      setattr(func, '_call_history', call_history)

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
      current_time = time.time()
      window_count = sum(ts > current_time - seconds for ts in call_history)
      if window_count >= count:
        # We need to sleep until the relevant call is outside the window.  This
        # should only ever be the the first entry in call_history, but if we
        # somehow ended up with extra calls in the window, this recovers.
        time.sleep(call_history[window_count - count] - current_time + seconds)
      # Append this call, deque will automatically trim old calls using maxlen.
      call_history.append(time.time())
      return func(*args, **kwargs)
    return _wrapper
  return decorator



# Copyright 2017 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""OpenHTF plug for serial port.

Allows for writing out to a serial port.
"""

import logging
import threading

try:
  import serial
except ImportError:
  logging.error('Failed to import pyserial. Please install the `serial_collection_plug` extra, '
                'e.g. via `pip install openhtf[serial_collection_plug]`.')
  raise

import openhtf
from openhtf.util import conf


conf.declare(
    'serial_collection_port',
    description='Port on which to collect serial data.',
    default_value='/dev/ttyACM0')
conf.declare(
    'serial_collection_baud',
    description='Baud rate for serial data collection.',
    default_value=115200)


class SerialCollectionPlug(openhtf.plugs.BasePlug):
  """Plug that collects data from a serial port.

  Spawns a thread that will open the configured serial port, continuously
  poll the port for data, and write the data to the destination file as it is
  received. If any serial errors are encountered during the lifetime of the
  polling thread, data collection stops and an error message is logged.
  Otherwise, data collection stops and the serial port is closed when
  stop_collection() is called.
  """
  # Serial library can raise these exceptions
  SERIAL_EXCEPTIONS = (serial.SerialException, ValueError)

  @conf.inject_positional_args
  def __init__(self, serial_collection_port, serial_collection_baud):
    super(SerialCollectionPlug, self).__init__()
    # Instantiate the port with no name, then add the name, so it won't be
    # opened until the collection context is entered.
    self._serial = serial.Serial(port=None,
                                 baudrate=serial_collection_baud,
                                 timeout=1)
    self._serial.port = serial_collection_port
    self._collect = False
    self._collection_thread = None

  def start_collection(self, dest):
    def _poll():
      try:
        with open(dest, 'w+') as outfile:
          while self._collect:
            data = self._serial.readline().decode()
            outfile.write(data)
      except self.SERIAL_EXCEPTIONS:
        self.logger.error('Serial port error. Stopping data collection.',
                          exc_info=True)

    self._collect = True
    self._collection_thread = threading.Thread(target=_poll)
    self._collection_thread.daemon = True
    self.logger.debug(
        'Starting serial data collection on port %s.' % self._serial.port)
    self._serial.open()
    self._collection_thread.start()

  @property
  def is_collecting(self):
    if self._collection_thread is not None:
      return self._collection_thread.is_alive()
    return False

  def stop_collection(self):
    if not self.is_collecting:
      self.logger.warning(
          'Data collection was not running, cannot be stopped.')
      return
    self._collect = False
    self._collection_thread.join()
    self._serial.close()
